#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import gzip
import io
from unittest import mock

import pandas as pd
import polars as pl
import pytest

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.models import Connection
from airflow.providers.amazon.aws.transfers.sql_to_s3 import SqlToS3Operator
from airflow.providers.common.compat.sdk import AirflowException


class TestSqlToS3Operator:
    @pytest.mark.parametrize(
        ("file_format", "dtype_backend", "df_kwargs", "expected_key_suffix"),
        [
            ("csv", "numpy_nullable", {"index": False, "header": False}, ".csv"),
            ("csv", "pyarrow", {"index": False, "header": False}, ".csv"),
            ("parquet", "numpy_nullable", {}, ".parquet"),
            ("parquet", "pyarrow", {}, ".parquet"),
            (
                "json",
                None,
                {"date_format": "iso", "lines": True, "orient": "records"},
                ".json",
            ),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_formats(self, mock_s3_hook, file_format, dtype_backend, df_kwargs, expected_key_suffix):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        read_df_kwargs = {"dtype_backend": dtype_backend} if dtype_backend else {}

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            file_format=file_format,
            replace=True,
            read_kwargs=read_df_kwargs,
            df_kwargs=df_kwargs,
            dag=None,
        )
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)

        expected_df_kwargs = {
            "sql": query,
            "parameters": None,
            "df_type": "pandas",
        }
        if dtype_backend:
            expected_df_kwargs["dtype_backend"] = dtype_backend

        get_df_mock.assert_called_once_with(**expected_df_kwargs)

        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)

        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj,
            key=f"{s3_key}{expected_key_suffix}",
            bucket_name=s3_bucket,
            replace=True,
        )

    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_gzip_with_bytesio(self, mock_s3_hook):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key.csv.gz"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            df_kwargs={"index": False, "compression": "gzip"},
            dag=None,
        )
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)
        get_df_mock.assert_called_once_with(sql=query, parameters=None, df_type="pandas")
        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)
        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj, key=s3_key, bucket_name=s3_bucket, replace=True
        )

        file_obj.seek(0)
        with gzip.GzipFile(fileobj=file_obj, mode="rb") as gz:
            decompressed_buf = io.BytesIO(gz.read())
        decompressed_buf.seek(0)
        read_df = pd.read_csv(decompressed_buf, dtype={"a": str, "b": str})
        assert read_df.equals(test_df)

    @pytest.mark.parametrize(
        "params",
        [
            pytest.param({"file_format": "csv", "null_string_result": None}, id="with-csv"),
            pytest.param({"file_format": "parquet", "null_string_result": "None"}, id="with-parquet"),
        ],
    )
    def test_fix_dtypes(self, params):
        op = SqlToS3Operator(
            query="query",
            s3_bucket="s3_bucket",
            s3_key="s3_key",
            task_id="task_id",
            sql_conn_id="mysql_conn_id",
            file_format=params["file_format"],
        )
        dirty_df = pd.DataFrame({"strings": ["a", "b", None], "ints": [1, 2, None]})
        op._fix_dtypes(df=dirty_df, file_format=op.file_format)
        assert dirty_df["strings"].values[2] == params["null_string_result"]
        assert dirty_df["ints"].dtype.kind == "i"

    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_fix_dtypes_not_called(self, mock_s3_hook):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])

        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            read_kwargs={"dtype_backend": "pyarrow"},
            file_format="parquet",
            replace=True,
            dag=None,
        )

        op._get_hook = mock_dbapi_hook

        with mock.patch.object(SqlToS3Operator, "_fix_dtypes") as mock_fix_dtypes:
            op.execute(None)

        mock_fix_dtypes.assert_not_called()

    def test_invalid_file_format(self):
        with pytest.raises(AirflowException):
            SqlToS3Operator(
                query="query",
                s3_bucket="bucket",
                s3_key="key",
                sql_conn_id="mysql_conn_id",
                task_id="task_id",
                file_format="invalid_format",
                dag=None,
            )

    def test_with_groupby_kwarg(self):
        """
        Test operator when the groupby_kwargs is specified
        """
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            df_kwargs={"index": False, "header": False},
            groupby_kwargs={"by": "Team"},
            dag=None,
        )
        example = {
            "Team": ["Australia", "Australia", "India", "India"],
            "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
            "Runs": [345, 490, 672, 560],
        }

        df = pd.DataFrame(example)
        data = []
        for group_name, df in op._partition_dataframe(df):
            data.append((group_name, df))
        data.sort(key=lambda d: d[0])
        team, df = data[0]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["Australia", "Australia"],
                    "Player": ["Ricky", "David Warner"],
                    "Runs": [345, 490],
                }
            )
        )
        team, df = data[1]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["India", "India"],
                    "Player": ["Virat Kohli", "Rohit Sharma"],
                    "Runs": [672, 560],
                }
            )
        )

    def test_without_groupby_kwarg(self):
        """
        Test operator when the groupby_kwargs is not specified
        """
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            df_kwargs={"index": False, "header": False},
            dag=None,
        )
        example = {
            "Team": ["Australia", "Australia", "India", "India"],
            "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
            "Runs": [345, 490, 672, 560],
        }

        df = pd.DataFrame(example)
        data = []
        for group_name, df in op._partition_dataframe(df):
            data.append((group_name, df))

        assert len(data) == 1
        team, df = data[0]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["Australia", "Australia", "India", "India"],
                    "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
                    "Runs": [345, 490, 672, 560],
                }
            )
        )

    def test_with_max_rows_per_file(self):
        """
        Test operator when the max_rows_per_file is specified
        """
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            replace=True,
            df_kwargs={"index": False, "header": False},
            max_rows_per_file=3,
            dag=None,
        )
        example = {
            "Team": ["Australia", "Australia", "India", "India"],
            "Player": ["Ricky", "David Warner", "Virat Kohli", "Rohit Sharma"],
            "Runs": [345, 490, 672, 560],
        }

        df = pd.DataFrame(example)
        data = []
        for group_name, df in op._partition_dataframe(df):
            data.append((group_name, df))
        data.sort(key=lambda d: d[0])
        team, df = data[0]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["Australia", "Australia", "India"],
                    "Player": ["Ricky", "David Warner", "Virat Kohli"],
                    "Runs": [345, 490, 672],
                }
            )
        )
        team, df = data[1]
        assert df.equals(
            pd.DataFrame(
                {
                    "Team": ["India"],
                    "Player": ["Rohit Sharma"],
                    "Runs": [560],
                }
            )
        )

    @mock.patch("airflow.providers.common.sql.operators.sql.BaseHook.get_connection")
    def test_hook_params(self, mock_get_conn):
        mock_get_conn.return_value = Connection(conn_id="postgres_test", conn_type="postgres")
        op = SqlToS3Operator(
            query="query",
            s3_bucket="bucket",
            s3_key="key",
            sql_conn_id="postgres_test",
            task_id="task_id",
            sql_hook_params={
                "log_sql": False,
            },
            dag=None,
        )
        hook = op._get_hook()
        assert hook.log_sql == op.sql_hook_params["log_sql"]

    @pytest.mark.parametrize(
        ("df_type_param", "expected_df_type"),
        [
            pytest.param("polars", "polars", id="with-polars"),
            pytest.param("pandas", "pandas", id="with-pandas"),
            pytest.param(None, "pandas", id="with-default"),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_execute_with_df_type(self, mock_s3_hook, df_type_param, expected_df_type):
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key.csv"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        kwargs = {
            "query": query,
            "s3_bucket": s3_bucket,
            "s3_key": s3_key,
            "sql_conn_id": "mysql_conn_id",
            "aws_conn_id": "aws_conn_id",
            "task_id": "task_id",
            "replace": True,
            "dag": None,
        }
        if df_type_param is not None:
            kwargs["df_type"] = df_type_param

        op = SqlToS3Operator(**kwargs)
        op._get_hook = mock_dbapi_hook
        op.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id="aws_conn_id", verify=None)
        get_df_mock.assert_called_once_with(sql=query, parameters=None, df_type=expected_df_type)
        file_obj = mock_s3_hook.return_value.load_file_obj.call_args[1]["file_obj"]
        assert isinstance(file_obj, io.BytesIO)
        mock_s3_hook.return_value.load_file_obj.assert_called_once_with(
            file_obj=file_obj, key=s3_key, bucket_name=s3_bucket, replace=True
        )

    @pytest.mark.parametrize(
        ("df_type", "input_df_creator"),
        [
            pytest.param(
                "pandas",
                lambda: pd.DataFrame({"category": ["A", "A", "B", "B"], "value": [1, 2, 3, 4]}),
                id="with-pandas-dataframe",
            ),
            pytest.param(
                "polars",
                lambda: pytest.importorskip("polars").DataFrame(
                    {"category": ["A", "A", "B", "B"], "value": [1, 2, 3, 4]}
                ),
                id="with-polars-dataframe",
            ),
        ],
    )
    def test_partition_dataframe(self, df_type, input_df_creator):
        """Test that _partition_dataframe works with both pandas and polars DataFrames."""
        op = SqlToS3Operator(
            query="query",
            s3_bucket="bucket",
            s3_key="key",
            sql_conn_id="mysql_conn_id",
            task_id="task_id",
            df_type=df_type,
            groupby_kwargs={"by": "category"},
        )

        input_df = input_df_creator()
        partitions = list(op._partition_dataframe(input_df))

        assert len(partitions) == 2
        for group_name, df in partitions:
            if df_type == "polars":
                assert isinstance(df, pl.DataFrame)
            else:
                assert isinstance(df, pd.DataFrame)
            assert group_name in ["A", "B"]

    @pytest.mark.parametrize(
        ("kwargs", "expected_warning", "expected_error", "expected_read_kwargs"),
        [
            pytest.param(
                {"read_pd_kwargs": {"dtype_backend": "pyarrow"}},
                "The 'read_pd_kwargs' parameter is deprecated",
                None,
                {"dtype_backend": "pyarrow"},
                id="deprecated-read-pd-kwargs-warning",
            ),
            pytest.param(
                {
                    "read_kwargs": {"dtype_backend": "pyarrow"},
                    "read_pd_kwargs": {"dtype_backend": "numpy_nullable"},
                },
                "The 'read_pd_kwargs' parameter is deprecated",
                None,
                {"dtype_backend": "pyarrow"},
                id="read-kwargs-priority-over-deprecated",
            ),
            pytest.param(
                {"max_rows_per_file": 2, "groupby_kwargs": {"by": "category"}},
                None,
                "can not be both specified",
                None,
                id="max-rows-groupby-conflict-error",
            ),
            pytest.param(
                {"pd_kwargs": {"index": False}},
                "The 'pd_kwargs' parameter is deprecated",
                None,
                None,
                id="deprecated-pd-kwargs-warning",
            ),
            pytest.param(
                {"df_kwargs": {"index": False}, "pd_kwargs": {"header": False}},
                "The 'pd_kwargs' parameter is deprecated",
                None,
                None,
                id="df-kwargs-priority-over-deprecated",
            ),
        ],
    )
    def test_parameter_validation(self, kwargs, expected_warning, expected_error, expected_read_kwargs):
        """Test parameter validation and deprecation warnings."""
        base_kwargs = {
            "query": "query",
            "s3_bucket": "bucket",
            "s3_key": "key",
            "sql_conn_id": "mysql_conn_id",
            "task_id": "task_id",
        }
        base_kwargs.update(kwargs)

        if expected_error:
            with pytest.raises(AirflowException, match=expected_error):
                SqlToS3Operator(**base_kwargs)
        elif expected_warning:
            with pytest.warns(AirflowProviderDeprecationWarning, match=expected_warning):
                op = SqlToS3Operator(**base_kwargs)
            if expected_read_kwargs:
                assert op.read_kwargs == expected_read_kwargs
        else:
            op = SqlToS3Operator(**base_kwargs)
            if expected_read_kwargs:
                assert op.read_kwargs == expected_read_kwargs

    @pytest.mark.parametrize(
        ("df_type", "should_call_fix_dtypes"),
        [
            pytest.param("pandas", True, id="pandas-calls-fix-dtypes"),
            pytest.param("polars", False, id="polars-skips-fix-dtypes"),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    def test_fix_dtypes_behavior_by_df_type(self, mock_s3_hook, df_type, should_call_fix_dtypes):
        """Test that _fix_dtypes is called/not called based on df_type."""
        query = "query"
        s3_bucket = "bucket"
        s3_key = "key"

        mock_dbapi_hook = mock.Mock()
        test_df = pd.DataFrame({"a": "1", "b": "2"}, index=[0, 1])
        get_df_mock = mock_dbapi_hook.return_value.get_df
        get_df_mock.return_value = test_df

        op = SqlToS3Operator(
            query=query,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="mysql_conn_id",
            aws_conn_id="aws_conn_id",
            task_id="task_id",
            df_type=df_type,
            replace=True,
            dag=None,
        )
        op._get_hook = mock_dbapi_hook

        with mock.patch.object(SqlToS3Operator, "_fix_dtypes") as mock_fix_dtypes:
            op.execute(None)

        if should_call_fix_dtypes:
            mock_fix_dtypes.assert_called_once()
        else:
            mock_fix_dtypes.assert_not_called()

    @pytest.mark.parametrize(
        ("kwargs", "expected_warning", "expected_read_kwargs", "expected_df_kwargs"),
        [
            pytest.param(
                {
                    "read_kwargs": {"dtype_backend": "pyarrow"},
                    "read_pd_kwargs": {"dtype_backend": "numpy_nullable"},
                },
                "The 'read_pd_kwargs' parameter is deprecated",
                {"dtype_backend": "pyarrow"},
                {},
                id="read-kwargs-priority-over-deprecated",
            ),
            pytest.param(
                {"read_pd_kwargs": {"dtype_backend": "numpy_nullable"}},
                "The 'read_pd_kwargs' parameter is deprecated",
                {"dtype_backend": "numpy_nullable"},
                {},
                id="read-pd-kwargs-used-when-read-kwargs-none",
            ),
            pytest.param(
                {
                    "df_kwargs": {"index": False},
                    "pd_kwargs": {"header": False},
                },
                "The 'pd_kwargs' parameter is deprecated",
                {},
                {"index": False},
                id="df-kwargs-priority-over-deprecated",
            ),
            pytest.param(
                {"pd_kwargs": {"header": False}},
                "The 'pd_kwargs' parameter is deprecated",
                {},
                {"header": False},
                id="pd-kwargs-used-when-df-kwargs-none",
            ),
        ],
    )
    def test_deprecated_kwargs_priority_behavior(
        self, kwargs, expected_warning, expected_read_kwargs, expected_df_kwargs
    ):
        """Test priority behavior and deprecation warnings for deprecated parameters."""
        base_kwargs = {
            "query": "query",
            "s3_bucket": "bucket",
            "s3_key": "key",
            "sql_conn_id": "mysql_conn_id",
            "task_id": "task_id",
        }
        base_kwargs.update(kwargs)

        with pytest.warns(AirflowProviderDeprecationWarning, match=expected_warning):
            op = SqlToS3Operator(**base_kwargs)

        assert op.read_kwargs == expected_read_kwargs
        assert op.df_kwargs == expected_df_kwargs

    @pytest.mark.parametrize(
        ("fmt", "df_kwargs", "expected_key"),
        [
            ("csv", {"compression": "gzip", "index": False}, "data.csv.gz"),
            ("csv", {"index": False}, "data.csv"),
            ("json", {"compression": "gzip"}, "data.json.gz"),
            ("json", {}, "data.json"),
            ("parquet", {"compression": "gzip"}, "data.parquet"),
            ("parquet", {}, "data.parquet"),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    @mock.patch("airflow.providers.common.sql.hooks.sql.DbApiHook")
    def test_file_format_handling(self, mock_dbapi_hook, mock_s3_hook, fmt, df_kwargs, expected_key):
        s3_bucket = "bucket"
        s3_key = "data." + fmt
        test_df = pd.DataFrame({"x": [1, 2]})
        mock_dbapi_hook.return_value.get_df.return_value = test_df

        op = SqlToS3Operator(
            query="SELECT * FROM test",
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="sqlite_conn",
            aws_conn_id="aws_default",
            task_id="task_id",
            file_format=fmt,
            df_kwargs=df_kwargs,
            replace=True,
            dag=None,
        )
        op._get_hook = lambda: mock_dbapi_hook.return_value
        op.execute(context=None)

        uploaded_key = mock_s3_hook.return_value.load_file_obj.call_args[1]["key"]
        assert uploaded_key == expected_key

    @pytest.mark.parametrize(
        ("file_format", "df_kwargs", "expected_suffix"),
        [
            ("csv", {"compression": "gzip", "index": False}, ".csv.gz"),
            ("csv", {"index": False}, ".csv"),
            ("json", {"compression": "gzip"}, ".json.gz"),
            ("json", {}, ".json"),
            ("parquet", {"compression": "gzip"}, ".parquet"),
            ("parquet", {}, ".parquet"),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.transfers.sql_to_s3.S3Hook")
    @mock.patch("airflow.providers.common.sql.hooks.sql.DbApiHook")
    def test_file_format_handling_with_groupby(
        self, mock_dbapi_hook, mock_s3_hook, file_format, df_kwargs, expected_suffix
    ):
        s3_bucket = "bucket"
        s3_key = "data"

        # Input DataFrame with groups
        test_data = pd.DataFrame(
            {"x": [1, 2, 3, 4, 5, 6], "group": ["group1", "group1", "group2", "group2", "group3", "group4"]}
        )

        mock_dbapi_hook.return_value.get_df.return_value = test_data

        op = SqlToS3Operator(
            query="SELECT * FROM test",
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            sql_conn_id="sqlite_conn",
            aws_conn_id="aws_default",
            task_id="task_id",
            file_format=file_format,
            df_kwargs=df_kwargs,
            groupby_kwargs={"by": "group"},
            replace=True,
            dag=None,
        )

        op._get_hook = lambda: mock_dbapi_hook.return_value
        op.execute(context=None)

        expected_groups = test_data["group"].unique()
        assert mock_s3_hook.return_value.load_file_obj.call_count == len(expected_groups)

        called_keys = [call.kwargs["key"] for call in mock_s3_hook.return_value.load_file_obj.call_args_list]

        for group in expected_groups:
            expected_key = f"{s3_key}_{group}{expected_suffix}"
            assert expected_key in called_keys, f"Missing expected key: {expected_key}"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest import mock

import pytest

from airflow.models import Connection
from airflow.providers.amazon.aws.hooks.redshift_sql import RedshiftSQLHook
from airflow.providers.amazon.version_compat import NOTSET
from airflow.providers.common.compat.sdk import AirflowException

LOGIN_USER = "login"
LOGIN_PASSWORD = "password"
LOGIN_HOST = "host"
LOGIN_PORT = 5439
LOGIN_SCHEMA = "dev"
MOCK_REGION_NAME = "eu-north-1"


class TestRedshiftSQLHookConn:
    def setup_method(self):
        self.connection = Connection(
            conn_type="redshift",
            login=LOGIN_USER,
            password=LOGIN_PASSWORD,
            host=LOGIN_HOST,
            port=LOGIN_PORT,
            schema=LOGIN_SCHEMA,
        )

        self.db_hook = RedshiftSQLHook()
        self.db_hook.get_connection = mock.Mock()
        self.db_hook.get_connection.return_value = self.connection

    def test_get_uri(self):
        db_uri = self.db_hook.get_uri()
        expected = "postgresql://login:password@host:5439/dev"
        assert db_uri == expected

    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.redshift_connector.connect")
    def test_get_conn(self, mock_connect):
        self.db_hook.get_conn()
        mock_connect.assert_called_once_with(
            user="login", password="password", host="host", port=5439, database="dev"
        )

    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.redshift_connector.connect")
    def test_get_conn_extra(self, mock_connect):
        self.connection.extra = json.dumps(
            {
                "iam": False,
                "cluster_identifier": "my-test-cluster",
                "profile": "default",
            }
        )
        self.db_hook.get_conn()
        mock_connect.assert_called_once_with(
            user=LOGIN_USER,
            password=LOGIN_PASSWORD,
            host=LOGIN_HOST,
            port=LOGIN_PORT,
            cluster_identifier="my-test-cluster",
            profile="default",
            database=LOGIN_SCHEMA,
            iam=False,
        )

    @mock.patch("airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook.conn")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.redshift_connector.connect")
    @pytest.mark.parametrize("aws_conn_id", [NOTSET, None, "mock_aws_conn"])
    def test_get_conn_iam(self, mock_connect, mock_aws_hook_conn, aws_conn_id):
        mock_conn_extra = {"iam": True, "profile": "default", "cluster_identifier": "my-test-cluster"}
        if aws_conn_id is not NOTSET:
            self.db_hook.aws_conn_id = aws_conn_id
        self.connection.extra = json.dumps(mock_conn_extra)

        mock_db_user = f"IAM:{self.connection.login}"
        mock_db_pass = "aws_token"

        # Mock AWS Connection
        mock_aws_hook_conn.get_cluster_credentials.return_value = {
            "DbPassword": mock_db_pass,
            "DbUser": mock_db_user,
        }

        self.db_hook.get_conn()

        # Check boto3 'redshift' client method `get_cluster_credentials` call args
        mock_aws_hook_conn.get_cluster_credentials.assert_called_once_with(
            DbUser=LOGIN_USER,
            DbName=LOGIN_SCHEMA,
            ClusterIdentifier="my-test-cluster",
            AutoCreate=False,
        )

        mock_connect.assert_called_once_with(
            user=mock_db_user,
            password=mock_db_pass,
            host=LOGIN_HOST,
            port=LOGIN_PORT,
            cluster_identifier="my-test-cluster",
            profile="default",
            database=LOGIN_SCHEMA,
            iam=True,
        )

    @mock.patch("airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook.conn")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.redshift_connector.connect")
    @pytest.mark.parametrize("aws_conn_id", [NOTSET, None, "mock_aws_conn"])
    def test_get_conn_iam_serverless_redshift(self, mock_connect, mock_aws_hook_conn, aws_conn_id):
        mock_work_group = "my-test-workgroup"
        mock_conn_extra = {
            "iam": True,
            "is_serverless": True,
            "profile": "default",
            "serverless_work_group": mock_work_group,
        }
        if aws_conn_id is not NOTSET:
            self.db_hook.aws_conn_id = aws_conn_id
        self.connection.extra = json.dumps(mock_conn_extra)

        mock_db_user = f"IAM:{self.connection.login}"
        mock_db_pass = "aws_token"

        # Mock AWS Connection
        mock_aws_hook_conn.get_credentials.return_value = {
            "dbPassword": mock_db_pass,
            "dbUser": mock_db_user,
        }

        self.db_hook.get_conn()

        # Check boto3 'redshift' client method `get_cluster_credentials` call args
        mock_aws_hook_conn.get_credentials.assert_called_once_with(
            dbName=LOGIN_SCHEMA,
            workgroupName=mock_work_group,
            durationSeconds=3600,
        )

        mock_connect.assert_called_once_with(
            user=mock_db_user,
            password=mock_db_pass,
            host=LOGIN_HOST,
            port=LOGIN_PORT,
            serverless_work_group=mock_work_group,
            profile="default",
            database=LOGIN_SCHEMA,
            iam=True,
            is_serverless=True,
        )

    @pytest.mark.parametrize(
        ("conn_params", "conn_extra", "expected_call_args"),
        [
            ({}, {}, {}),
            ({"login": "test"}, {}, {"user": "test"}),
            ({}, {"user": "test"}, {"user": "test"}),
            ({"login": "original"}, {"user": "overridden"}, {"user": "overridden"}),
            ({"login": "test1"}, {"password": "test2"}, {"user": "test1", "password": "test2"}),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.redshift_connector.connect")
    def test_get_conn_overrides_correctly(self, mock_connect, conn_params, conn_extra, expected_call_args):
        with mock.patch(
            "airflow.providers.amazon.aws.hooks.redshift_sql.RedshiftSQLHook.conn",
            Connection(conn_type="redshift", extra=conn_extra, **conn_params),
        ):
            self.db_hook.get_conn()
            mock_connect.assert_called_once_with(**expected_call_args)

    @pytest.mark.parametrize(
        ("connection_host", "connection_extra", "expected_cluster_identifier", "expected_exception_msg"),
        [
            # test without a connection host and without a cluster_identifier in connection extra
            (None, {"iam": True}, None, "Please set cluster_identifier or host in redshift connection."),
            # test without a connection host but with a cluster_identifier in connection extra
            (
                None,
                {"iam": True, "cluster_identifier": "cluster_identifier_from_extra"},
                "cluster_identifier_from_extra",
                None,
            ),
            # test with a connection host and without a cluster_identifier in connection extra
            ("cluster_identifier_from_host.x.y", {"iam": True}, "cluster_identifier_from_host", None),
            # test with both connection host and cluster_identifier in connection extra
            (
                "cluster_identifier_from_host.x.y",
                {"iam": True, "cluster_identifier": "cluster_identifier_from_extra"},
                "cluster_identifier_from_extra",
                None,
            ),
        ],
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook.conn")
    @mock.patch("airflow.providers.amazon.aws.hooks.redshift_sql.redshift_connector.connect")
    def test_get_iam_token(
        self,
        mock_connect,
        mock_aws_hook_conn,
        connection_host,
        connection_extra,
        expected_cluster_identifier,
        expected_exception_msg,
    ):
        self.connection.host = connection_host
        self.connection.extra = json.dumps(connection_extra)

        mock_db_user = f"IAM:{self.connection.login}"
        mock_db_pass = "aws_token"

        # Mock AWS Connection
        mock_aws_hook_conn.get_cluster_credentials.return_value = {
            "DbPassword": mock_db_pass,
            "DbUser": mock_db_user,
        }
        if expected_exception_msg is not None:
            with pytest.raises(AirflowException, match=expected_exception_msg):
                self.db_hook.get_conn()
        else:
            self.db_hook.get_conn()
            mock_aws_hook_conn.get_cluster_credentials.assert_called_once_with(
                DbUser=LOGIN_USER,
                DbName=LOGIN_SCHEMA,
                ClusterIdentifier=expected_cluster_identifier,
                AutoCreate=False,
            )

    @mock.patch.dict("os.environ", AIRFLOW_CONN_AWS_DEFAULT=f"aws://?region_name={MOCK_REGION_NAME}")
    @pytest.mark.parametrize(
        ("connection_host", "connection_extra", "expected_identity"),
        [
            # test without a connection host but with a cluster_identifier in connection extra
            (
                None,
                {"iam": True, "cluster_identifier": "cluster_identifier_from_extra"},
                f"cluster_identifier_from_extra.{MOCK_REGION_NAME}",
            ),
            # test with a connection host and without a cluster_identifier in connection extra
            (
                "cluster_identifier_from_host.id.my_region.redshift.amazonaws.com",
                {"iam": True},
                "cluster_identifier_from_host.my_region",
            ),
            # test with both connection host and cluster_identifier in connection extra
            (
                "cluster_identifier_from_host.x.y",
                {"iam": True, "cluster_identifier": "cluster_identifier_from_extra"},
                f"cluster_identifier_from_extra.{MOCK_REGION_NAME}",
            ),
            # test when hostname doesn't match pattern
            (
                "1.2.3.4",
                {},
                "1.2.3.4",
            ),
        ],
    )
    def test_get_openlineage_redshift_authority_part(
        self,
        connection_host,
        connection_extra,
        expected_identity,
    ):
        self.connection.host = connection_host
        self.connection.extra = json.dumps(connection_extra)

        assert f"{expected_identity}:{LOGIN_PORT}" == self.db_hook._get_openlineage_redshift_authority_part(
            self.connection
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains the unit tests for the helper methods included in the Amazon System Tests found at
providers/amazon/tests/system/amazon/aws/utils/__init__.py
"""

from __future__ import annotations

import os
import sys
from io import StringIO
from unittest.mock import patch

import pytest
from moto import mock_aws

from system.amazon.aws import utils
from system.amazon.aws.utils import (
    DEFAULT_ENV_ID_LEN,
    DEFAULT_ENV_ID_PREFIX,
    ENV_ID_ENVIRON_KEY,
    INVALID_ENV_ID_MSG,
    LOWERCASE_ENV_ID_MSG,
    NO_VALUE_MSG,
    _validate_env_id,
    set_env_id,
)

TEST_NAME: str = "example_test"
ANY_STR: str = "any"

ENV_VALUE: str = "foo"
SSM_VALUE: str = "bar"
DEFAULT_VALUE: str = "baz"


@pytest.fixture(autouse=True)
def provide_test_name():
    with patch.object(utils, "_get_test_name", return_value=TEST_NAME) as name:
        yield name


@mock_aws
class TestAmazonSystemTestHelpers:
    FETCH_VARIABLE_TEST_CASES = [
        # Format is:
        # (Environment Variable value, Fetched SSM value, Provided Default value, Expected Result)
        (ENV_VALUE, SSM_VALUE, DEFAULT_VALUE, ENV_VALUE),
        (ENV_VALUE, SSM_VALUE, None, ENV_VALUE),
        (ENV_VALUE, None, DEFAULT_VALUE, ENV_VALUE),
        (ENV_VALUE, None, None, ENV_VALUE),
        (None, SSM_VALUE, DEFAULT_VALUE, SSM_VALUE),
        (None, SSM_VALUE, None, SSM_VALUE),
        (None, None, DEFAULT_VALUE, DEFAULT_VALUE),
        # For the (None, None, None ) test case, see: test_fetch_variable_no_value_found_raises_exception
    ]

    @pytest.mark.parametrize(
        ("env_value", "ssm_value", "default_value", "expected_result"), FETCH_VARIABLE_TEST_CASES
    )
    @patch.object(os, "getenv")
    def test_fetch_variable_success(
        self, mock_getenv, env_value, ssm_value, default_value, expected_result
    ) -> None:
        mock_getenv.return_value = env_value or ssm_value

        utils._fetch_from_ssm.cache_clear()

        result = (
            utils.fetch_variable(key="some_key", test_name=TEST_NAME, default_value=default_value)
            if default_value
            else utils.fetch_variable(key=ANY_STR, test_name=TEST_NAME)
        )

        utils._fetch_from_ssm.cache_clear()
        assert result == expected_result

    def test_fetch_variable_no_value_found_raises_exception(self):
        # This would be the (None, None, None) test case from above.
        with pytest.raises(ValueError, match=NO_VALUE_MSG.format(key=ANY_STR)):
            utils.fetch_variable(key=ANY_STR, test_name=TEST_NAME)

    ENV_ID_TEST_CASES = [
        # Happy Cases
        ("ABCD", True),
        ("AbCd", True),
        ("abcd", True),
        ("ab12", True),
        # Failure Cases
        # Must be alphanumeric
        ("not_alphanumeric", False),
        # Can not be empty
        ("", False),
        # Must start with a letter
        ("1234", False),
        ("12ab", False),
        ("12AB", False),
        ("12Ab", False),
    ]

    @pytest.mark.parametrize(("env_id", "is_valid"), ENV_ID_TEST_CASES)
    def test_validate_env_id_success(self, env_id, is_valid):
        if is_valid:
            captured_output = StringIO()
            sys.stdout = captured_output

            result = _validate_env_id(env_id)
            sys.stdout = sys.__stdout__

            assert result == env_id.lower()
            assert result.isalnum()
            if not result == env_id:
                assert LOWERCASE_ENV_ID_MSG in captured_output.getvalue()
        else:
            with pytest.raises(ValueError, match=INVALID_ENV_ID_MSG):
                _validate_env_id(env_id)

    def test_set_env_id_generates_if_required(self):
        # No environment variable nor SSM value has been found
        result = set_env_id(TEST_NAME)

        assert len(result) == DEFAULT_ENV_ID_LEN + len(DEFAULT_ENV_ID_PREFIX)
        assert result.isalnum()
        assert result.islower()

    def test_set_env_id_exports_environment_variable(self):
        env_id = set_env_id(TEST_NAME)

        assert os.environ[ENV_ID_ENVIRON_KEY] == env_id



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TypeVar
from unittest import mock

import boto3
import pytest
from slugify import slugify

from airflow.providers.amazon.aws.sensors.ecs import (
    EcsBaseSensor,
    EcsClusterStates,
    EcsClusterStateSensor,
    EcsHook,
    EcsTaskDefinitionStates,
    EcsTaskDefinitionStateSensor,
    EcsTaskStates,
    EcsTaskStateSensor,
)
from airflow.providers.amazon.version_compat import NOTSET
from airflow.providers.common.compat.sdk import AirflowException

from tests_common.test_utils.compat import timezone

_Operator = TypeVar("_Operator")
TEST_CLUSTER_NAME = "fake-cluster"
TEST_TASK_ARN = "arn:aws:ecs:us-east-1:012345678910:task/spam-egg"
TEST_TASK_DEFINITION_ARN = "arn:aws:ecs:us-east-1:012345678910:task-definition/foo-bar:42"

pytestmark = pytest.mark.db_test


class EcsBaseTestCase:
    @pytest.fixture(autouse=True)
    def _setup_test_cases(self, monkeypatch, request, create_task_instance_of_operator, session):
        self.dag_id = f"dag-{slugify(request.cls.__name__)}"
        self.task_id = f"task-{slugify(request.node.name, max_length=40)}"
        self.fake_client = boto3.client("ecs", region_name="eu-west-3")
        monkeypatch.setattr(EcsHook, "conn", self.fake_client)
        self.ti_maker = create_task_instance_of_operator
        self.session = session

    def create_rendered_task(self, operator_class: type[_Operator], **kwargs) -> _Operator:
        """
        Create operator from given class and render fields.

        This might help to prevent of unexpected behaviour in Jinja/task field serialisation
        """
        ti = self.ti_maker(
            operator_class,
            dag_id=self.dag_id,
            task_id=self.task_id,
            logical_date=timezone.datetime(2021, 12, 21),
            **kwargs,
        )
        self.session.add(ti)
        self.session.commit()
        return ti.render_templates()


class TestEcsBaseSensor(EcsBaseTestCase):
    @pytest.mark.parametrize("aws_conn_id", [None, NOTSET, "aws_test_conn"])
    @pytest.mark.parametrize("region_name", [None, NOTSET, "ca-central-1"])
    def test_initialise_operator(self, aws_conn_id, region_name):
        """Test sensor initialize."""
        op_kw = {"aws_conn_id": aws_conn_id, "region_name": region_name}
        op_kw = {k: v for k, v in op_kw.items() if v is not NOTSET}
        op = EcsBaseSensor(task_id="test_ecs_base", **op_kw)

        assert op.aws_conn_id == (aws_conn_id if aws_conn_id is not NOTSET else "aws_default")
        assert op.region_name == (region_name if region_name is not NOTSET else None)

    @pytest.mark.parametrize("aws_conn_id", [None, NOTSET, "aws_test_conn"])
    @pytest.mark.parametrize("region_name", [None, NOTSET, "ca-central-1"])
    def test_hook_and_client(self, aws_conn_id, region_name):
        """Test initialize ``EcsHook`` and ``boto3.client``."""
        op_kw = {"aws_conn_id": aws_conn_id, "region_name": region_name}
        op_kw = {k: v for k, v in op_kw.items() if v is not NOTSET}
        op = EcsBaseSensor(task_id="test_ecs_base_hook_client", **op_kw)

        hook = op.hook
        assert op.hook is hook, "'hook' property should be cached."
        assert isinstance(op.hook, EcsHook)

        client = op.client
        assert op.client is client, "'client' property should be cached."
        assert client is self.fake_client


class TestEcsClusterStateSensor(EcsBaseTestCase):
    @pytest.mark.parametrize(
        ("return_state", "expected"), [("ACTIVE", True), ("PROVISIONING", False), ("DEPROVISIONING", False)]
    )
    def test_default_values_poke(self, return_state, expected):
        task = self.create_rendered_task(EcsClusterStateSensor, cluster_name=TEST_CLUSTER_NAME)
        with mock.patch.object(task.hook, "get_cluster_state") as m:
            m.return_value = return_state
            assert task.poke({}) == expected
            m.assert_called_once_with(cluster_name=TEST_CLUSTER_NAME)

    @pytest.mark.parametrize("return_state", ["FAILED", "INACTIVE"])
    def test_default_values_terminal_state(self, return_state):
        task = self.create_rendered_task(EcsClusterStateSensor, cluster_name=TEST_CLUSTER_NAME)
        with mock.patch.object(task.hook, "get_cluster_state") as m:
            m.return_value = return_state
            with pytest.raises(AirflowException, match="Terminal state reached"):
                task.poke({})
            m.assert_called_once_with(cluster_name=TEST_CLUSTER_NAME)

    @pytest.mark.parametrize(
        ("target_state", "return_state", "expected"),
        [
            (EcsClusterStates.ACTIVE, "ACTIVE", True),
            (EcsClusterStates.ACTIVE, "DEPROVISIONING", False),
            (EcsClusterStates.DEPROVISIONING, "ACTIVE", False),
            (EcsClusterStates.DEPROVISIONING, "DEPROVISIONING", True),
        ],
    )
    def test_custom_values_poke(self, target_state, return_state, expected):
        task = self.create_rendered_task(
            EcsClusterStateSensor, cluster_name=TEST_CLUSTER_NAME, target_state=target_state
        )
        with mock.patch.object(task.hook, "get_cluster_state") as m:
            m.return_value = return_state
            assert task.poke({}) == expected
            m.assert_called_once_with(cluster_name=TEST_CLUSTER_NAME)

    @pytest.mark.parametrize(
        ("failure_states", "return_state"),
        [
            ({EcsClusterStates.ACTIVE}, "ACTIVE"),
            ({EcsClusterStates.PROVISIONING, EcsClusterStates.DEPROVISIONING}, "DEPROVISIONING"),
            ({EcsClusterStates.PROVISIONING, EcsClusterStates.DEPROVISIONING}, "PROVISIONING"),
        ],
    )
    def test_custom_values_terminal_state(self, failure_states, return_state):
        task = self.create_rendered_task(
            EcsClusterStateSensor,
            cluster_name=TEST_CLUSTER_NAME,
            target_state=EcsClusterStates.FAILED,
            failure_states=failure_states,
        )
        with mock.patch.object(task.hook, "get_cluster_state") as m:
            m.return_value = return_state
            with pytest.raises(AirflowException, match="Terminal state reached"):
                task.poke({})
            m.assert_called_once_with(cluster_name=TEST_CLUSTER_NAME)


class TestEcsTaskDefinitionStateSensor(EcsBaseTestCase):
    @pytest.mark.parametrize(
        ("return_state", "expected"), [("ACTIVE", True), ("INACTIVE", False), ("DELETE_IN_PROGRESS", False)]
    )
    def test_default_values_poke(self, return_state, expected):
        task = self.create_rendered_task(
            EcsTaskDefinitionStateSensor, task_definition=TEST_TASK_DEFINITION_ARN
        )
        with mock.patch.object(task.hook, "get_task_definition_state") as m:
            m.return_value = return_state
            assert task.poke({}) == expected
            m.assert_called_once_with(task_definition=TEST_TASK_DEFINITION_ARN)

    @pytest.mark.parametrize(
        ("target_state", "return_state", "expected"),
        [
            (EcsTaskDefinitionStates.INACTIVE, "ACTIVE", False),
            (EcsTaskDefinitionStates.INACTIVE, "INACTIVE", True),
            (EcsTaskDefinitionStates.ACTIVE, "INACTIVE", False),
            (EcsTaskDefinitionStates.ACTIVE, "ACTIVE", True),
        ],
    )
    def test_custom_values_poke(self, create_task_of_operator, target_state, return_state, expected):
        task = self.create_rendered_task(
            EcsTaskDefinitionStateSensor, task_definition=TEST_TASK_DEFINITION_ARN, target_state=target_state
        )
        with mock.patch.object(task.hook, "get_task_definition_state") as m:
            m.return_value = return_state
            assert task.poke({}) == expected
            m.assert_called_once_with(task_definition=TEST_TASK_DEFINITION_ARN)


class TestEcsTaskStateSensor(EcsBaseTestCase):
    @pytest.mark.parametrize(
        ("return_state", "expected"),
        [
            ("PROVISIONING", False),
            ("PENDING", False),
            ("ACTIVATING", False),
            ("RUNNING", True),
            ("DEACTIVATING", False),
            ("STOPPING", False),
            ("DEPROVISIONING", False),
            ("NONE", False),
        ],
    )
    def test_default_values_poke(self, return_state, expected):
        task = self.create_rendered_task(EcsTaskStateSensor, cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN)
        with mock.patch.object(task.hook, "get_task_state") as m:
            m.return_value = return_state
            assert task.poke({}) == expected
            m.assert_called_once_with(cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN)

    @pytest.mark.parametrize("return_state", ["STOPPED"])
    def test_default_values_terminal_state(self, return_state):
        task = self.create_rendered_task(EcsTaskStateSensor, cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN)
        with mock.patch.object(task.hook, "get_task_state") as m:
            m.return_value = return_state
            with pytest.raises(AirflowException, match="Terminal state reached"):
                task.poke({})
            m.assert_called_once_with(cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN)

    @pytest.mark.parametrize(
        ("target_state", "return_state", "expected"),
        [
            (EcsTaskStates.RUNNING, "RUNNING", True),
            (EcsTaskStates.DEACTIVATING, "DEACTIVATING", True),
            (EcsTaskStates.NONE, "PENDING", False),
            (EcsTaskStates.STOPPING, "NONE", False),
        ],
    )
    def test_custom_values_poke(self, target_state, return_state, expected):
        task = self.create_rendered_task(
            EcsTaskStateSensor, cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN, target_state=target_state
        )
        with mock.patch.object(task.hook, "get_task_state") as m:
            m.return_value = return_state
            assert task.poke({}) == expected
            m.assert_called_once_with(cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN)

    @pytest.mark.parametrize(
        ("failure_states", "return_state"),
        [
            ({EcsTaskStates.RUNNING}, "RUNNING"),
            ({EcsTaskStates.RUNNING, EcsTaskStates.DEACTIVATING}, "DEACTIVATING"),
            ({EcsTaskStates.RUNNING, EcsTaskStates.DEACTIVATING}, "RUNNING"),
        ],
    )
    def test_custom_values_terminal_state(self, failure_states, return_state):
        task = self.create_rendered_task(
            EcsTaskStateSensor,
            cluster=TEST_CLUSTER_NAME,
            task=TEST_TASK_ARN,
            target_state=EcsTaskStates.NONE,
            failure_states=failure_states,
        )
        with mock.patch.object(task.hook, "get_task_state") as m:
            m.return_value = return_state
            with pytest.raises(AirflowException, match="Terminal state reached"):
                task.poke({})
            m.assert_called_once_with(cluster=TEST_CLUSTER_NAME, task=TEST_TASK_ARN)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
import textwrap
import time
from datetime import datetime as dt, timedelta, timezone
from unittest import mock
from unittest.mock import ANY, call

import boto3
import pendulum
import pytest
import time_machine
from moto import mock_aws
from pydantic import TypeAdapter
from watchtower import CloudWatchLogHandler

from airflow.models import DAG, DagRun, TaskInstance
from airflow.providers.amazon.aws.hooks.logs import AwsLogsHook
from airflow.providers.amazon.aws.log.cloudwatch_task_handler import (
    CloudWatchRemoteLogIO,
    CloudwatchTaskHandler,
)
from airflow.providers.amazon.aws.utils import datetime_to_epoch_utc_ms
from airflow.utils.state import State
from airflow.utils.timezone import datetime

from tests_common.test_utils.compat import EmptyOperator
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.dag import sync_dag_to_db
from tests_common.test_utils.db import clear_db_dag_bundles, clear_db_dags, clear_db_runs
from tests_common.test_utils.taskinstance import create_task_instance
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_2_PLUS


def get_time_str(time_in_milliseconds):
    dt_time = dt.fromtimestamp(time_in_milliseconds / 1000.0, tz=timezone.utc)
    return dt_time.strftime("%Y-%m-%dT%H:%M:%SZ")


@pytest.fixture(autouse=True)
def logmock():
    with mock_aws():
        yield


# We only test this directly on Airflow 3
@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="This path only works on Airflow 3")
class TestCloudRemoteLogIO:
    # We use the cap_structlog so that our changes get reverted for us
    @pytest.fixture(autouse=True)
    def setup_tests(self, create_runtime_ti, tmp_path, monkeypatch):
        import structlog

        import airflow.logging_config
        import airflow.sdk.log
        from airflow.sdk import BaseOperator

        task = BaseOperator(task_id="task_1")
        self.ti = create_runtime_ti(task)

        self.remote_log_group = "log_group_name"
        self.region_name = "us-west-2"
        self.task_log_path = "dag_id=a/0:0.log"
        self.local_log_location = tmp_path / "local-cloudwatch-log-location"
        self.local_log_location.mkdir()
        # Create the local log file structure
        task_log_path_parts = self.task_log_path.split("/")
        dag_dir = self.local_log_location / task_log_path_parts[0]
        dag_dir.mkdir()
        task_log_file = dag_dir / task_log_path_parts[1]
        task_log_file.touch()

        # The subject under test
        self.subject = CloudWatchRemoteLogIO(
            base_log_folder=self.local_log_location,
            log_group_arn=f"arn:aws:logs:{self.region_name}:11111111:log-group:{self.remote_log_group}",
        )

        conn = boto3.client("logs", region_name=self.region_name)
        conn.create_log_group(logGroupName=self.remote_log_group)

        processors = structlog.get_config()["processors"]
        logger_factory = structlog.get_config()["logger_factory"]
        old_processors = processors.copy()

        try:
            # Modify `_Configuration.default_processors` set via `configure` but always
            # keep the list instance intact to not break references held by bound
            # loggers.

            # Set up the right chain of processors so the event looks like we want for our full test
            if AIRFLOW_V_3_2_PLUS:
                monkeypatch.setattr(
                    airflow.logging_config._ActiveLoggingConfig, "remote_task_log", self.subject
                )
            else:
                monkeypatch.setattr(airflow.logging_config, "REMOTE_TASK_LOG", self.subject)
            try:
                procs = airflow.sdk.log.logging_processors(colors=False, json_output=False)
            except TypeError:
                # Compat issue only comes up in the tests, not in the real code
                procs, _ = airflow.sdk.log.logging_processors(enable_pretty_log=False)
            processors.clear()
            processors.extend(procs)

            # Replace the last "output" one with a DropEvent one instead - else we get the output on stdout
            def drop(*args):
                raise structlog.DropEvent()

            processors[-1] = drop
            structlog.configure(
                processors=processors,
                # Create a logger factory and pass in the file path we want it to use
                # This is because we use the logger to determine the streamname/filepath
                # in the CloudWatchRemoteLogIO processor.
                logger_factory=structlog.PrintLoggerFactory(file=task_log_file.open("w+")),
            )
            yield
        finally:
            # remove LogCapture and restore original processors
            processors.clear()
            processors.extend(old_processors)
            structlog.configure(processors=old_processors, logger_factory=logger_factory)

    @time_machine.travel(datetime(2025, 3, 27, 21, 58, 1, 2345), tick=False)
    def test_log_message(self):
        # Use a context instead of a decorator on the test method because we need access to self to
        # get the path from the setup method.
        with conf_vars({("logging", "base_log_folder"): self.local_log_location.as_posix()}):
            import structlog

            log = structlog.get_logger()
            log.info("Hi", foo="bar")
            # We need to close in order to flush the logs etc.
            self.subject.close()

            # close call should only flush the logs and not set shutting_down to True
            assert self.subject.handler.shutting_down is False

            # Inside the Cloudwatch logger we swap colons for underscores since colons are not allowed in
            # stream names.
            stream_name = self.task_log_path.replace(":", "_")
            logs = self.subject.read(stream_name, self.ti)

            metadata, logs = logs

            assert metadata == [
                f"Reading remote log from Cloudwatch log_group: log_group_name log_stream: {stream_name}"
            ]
            assert logs == [
                '{"foo": "bar", "event": "Hi", "level": "info", "timestamp": "2025-03-27T21:58:01.002000+00:00"}\n'
            ]


@pytest.mark.db_test
class TestCloudwatchTaskHandler:
    def clear_db(self):
        if AIRFLOW_V_3_0_PLUS:
            clear_db_runs()
            clear_db_dags()
            clear_db_dag_bundles()

    @pytest.fixture(autouse=True)
    def setup(self, create_log_template, tmp_path_factory, session, testing_dag_bundle):
        # self.clear_db()
        with conf_vars({("logging", "remote_log_conn_id"): "aws_default"}):
            self.remote_log_group = "log_group_name"
            self.region_name = "us-west-2"
            self.local_log_location = str(tmp_path_factory.mktemp("local-cloudwatch-log-location"))
            if AIRFLOW_V_3_0_PLUS:
                create_log_template("{dag_id}/{task_id}/{logical_date}/{try_number}.log")
            else:
                create_log_template("{dag_id}/{task_id}/{execution_date}/{try_number}.log")
            self.cloudwatch_task_handler = CloudwatchTaskHandler(
                self.local_log_location,
                f"arn:aws:logs:{self.region_name}:11111111:log-group:{self.remote_log_group}",
            )

        date = datetime(2020, 1, 1)
        dag_id = "dag_for_testing_cloudwatch_task_handler"
        task_id = "task_for_testing_cloudwatch_log_handler"
        self.dag = DAG(dag_id=dag_id, schedule=None, start_date=date)
        task = EmptyOperator(task_id=task_id, dag=self.dag)
        if AIRFLOW_V_3_0_PLUS:
            sync_dag_to_db(self.dag)
            dag_run = DagRun(
                dag_id=self.dag.dag_id,
                logical_date=date,
                run_id="test",
                run_type="scheduled",
            )
        else:
            dag_run = DagRun(
                dag_id=self.dag.dag_id,
                execution_date=date,
                run_id="test",
                run_type="scheduled",
            )
        session.add(dag_run)
        session.commit()
        session.refresh(dag_run)

        if AIRFLOW_V_3_0_PLUS:
            from airflow.models.dag_version import DagVersion

            dag_version = DagVersion.get_latest_version(self.dag.dag_id, session=session)
            self.ti = create_task_instance(task=task, run_id=dag_run.run_id, dag_version_id=dag_version.id)
        else:
            self.ti = TaskInstance(task=task, run_id=dag_run.run_id)
        self.ti.dag_run = dag_run
        self.ti.try_number = 1
        self.ti.state = State.RUNNING
        session.add(self.ti)
        session.commit()

        self.remote_log_stream = (f"{dag_id}/{task_id}/{date.isoformat()}/{self.ti.try_number}.log").replace(
            ":", "_"
        )
        self.conn = boto3.client("logs", region_name=self.region_name)

        yield

        self.cloudwatch_task_handler.handler = None
        del self.cloudwatch_task_handler

        self.clear_db()

    def test_hook(self):
        assert isinstance(self.cloudwatch_task_handler.hook, AwsLogsHook)

    def test_handler(self):
        self.cloudwatch_task_handler.set_context(self.ti)
        assert isinstance(self.cloudwatch_task_handler.handler, CloudWatchLogHandler)

    def test_write(self):
        handler = self.cloudwatch_task_handler
        handler.set_context(self.ti)
        messages = [str(i) for i in range(10)]

        with mock.patch("watchtower.CloudWatchLogHandler.emit") as mock_emit:
            for message in messages:
                handler.handle(message)
            mock_emit.assert_has_calls([call(message) for message in messages])

    @time_machine.travel(datetime(2025, 3, 27, 21, 58, 1, 2345), tick=False)
    def test_read(self, monkeypatch):
        # Confirmed via AWS Support call:
        # CloudWatch events must be ordered chronologically otherwise
        # boto3 put_log_event API throws InvalidParameterException
        # (moto does not throw this exception)
        current_time = int(time.time()) * 1000
        generate_log_events(
            self.conn,
            self.remote_log_group,
            self.remote_log_stream,
            [
                {"timestamp": current_time - 2000, "message": "First"},
                {"timestamp": current_time - 1000, "message": "Second"},
                {"timestamp": current_time, "message": "Third"},
            ],
        )
        monkeypatch.setattr(
            self.cloudwatch_task_handler,
            "_read_from_logs_server",
            lambda ti, worker_log_rel_path: ([], []),
        )
        msg_template = textwrap.dedent("""
             INFO - ::group::Log message source details
            *** Reading remote log from Cloudwatch log_group: {} log_stream: {}
             INFO - ::endgroup::
            {}
        """)[1:][:-1]  # Strip off leading and trailing new lines, but not spaces

        logs, metadata = self.cloudwatch_task_handler.read(self.ti)
        if AIRFLOW_V_3_0_PLUS:
            from airflow.utils.log.file_task_handler import StructuredLogMessage

            logs = list(logs)
            results = TypeAdapter(list[StructuredLogMessage]).dump_python(logs)
            assert results[-4:] == [
                {"event": "::endgroup::", "timestamp": None},
                {
                    "event": "[2025-03-27T21:57:59Z] First",
                    "timestamp": pendulum.datetime(2025, 3, 27, 21, 57, 59),
                },
                {
                    "event": "[2025-03-27T21:58:00Z] Second",
                    "timestamp": pendulum.datetime(2025, 3, 27, 21, 58, 0),
                },
                {
                    "event": "[2025-03-27T21:58:01Z] Third",
                    "timestamp": pendulum.datetime(2025, 3, 27, 21, 58, 1),
                },
            ]
            assert metadata == {"end_of_log": False, "log_pos": 3}
        else:
            events = "\n".join(
                [
                    f"[{get_time_str(current_time - 2000)}] First",
                    f"[{get_time_str(current_time - 1000)}] Second",
                    f"[{get_time_str(current_time)}] Third",
                ]
            )
            assert logs == [
                [
                    (
                        "",
                        msg_template.format(self.remote_log_group, self.remote_log_stream, events),
                    )
                ]
            ]

    @pytest.mark.parametrize(
        ("end_date", "expected_end_time"),
        [
            (None, None),
            (
                datetime(2020, 1, 2),
                datetime_to_epoch_utc_ms(datetime(2020, 1, 2) + timedelta(seconds=30)),
            ),
        ],
    )
    @mock.patch.object(AwsLogsHook, "get_log_events")
    def test_get_cloudwatch_logs(self, mock_get_log_events, end_date, expected_end_time):
        self.ti.end_date = end_date
        self.cloudwatch_task_handler.io.get_cloudwatch_logs(self.remote_log_stream, self.ti)
        mock_get_log_events.assert_called_once_with(
            log_group=self.remote_log_group,
            log_stream_name=self.remote_log_stream,
            end_time=expected_end_time,
        )

    @pytest.mark.parametrize(
        ("conf_json_serialize", "expected_serialized_output"),
        [
            pytest.param(
                "airflow.providers.amazon.aws.log.cloudwatch_task_handler.json_serialize_legacy",
                '{"datetime": "2023-01-01T00:00:00+00:00", "customObject": null}',
                id="json-serialize-legacy",
            ),
            pytest.param(
                "airflow.providers.amazon.aws.log.cloudwatch_task_handler.json_serialize",
                '{"datetime": "2023-01-01T00:00:00+00:00", "customObject": "SomeCustomSerialization(...)"}',
                id="json-serialize",
            ),
            pytest.param(
                None,
                '{"datetime": "2023-01-01T00:00:00+00:00", "customObject": null}',
                id="not-set",
            ),
        ],
    )
    @mock.patch.object(AwsLogsHook, "get_log_events")
    def test_write_json_logs(self, mock_get_log_events, conf_json_serialize, expected_serialized_output):
        class ToSerialize:
            def __init__(self):
                pass

            def __repr__(self):
                return "SomeCustomSerialization(...)"

        with conf_vars({("aws", "cloudwatch_task_handler_json_serializer"): conf_json_serialize}):
            handler = self.cloudwatch_task_handler
            handler.set_context(self.ti)
            message = logging.LogRecord(
                name="test_log_record",
                level=logging.DEBUG,
                pathname="fake.path",
                lineno=42,
                args=None,
                exc_info=None,
                msg={
                    "datetime": datetime(2023, 1, 1),
                    "customObject": ToSerialize(),
                },
            )
            with (
                mock.patch("watchtower.threading.Thread"),
                mock.patch("watchtower.queue.Queue") as mq,
            ):
                handler.handle(message)
                mq.return_value.put.assert_called_once_with(
                    {"message": expected_serialized_output, "timestamp": ANY}
                )

    def test_close_prevents_duplicate_calls(self):
        with mock.patch("watchtower.CloudWatchLogHandler.close") as mock_log_handler_close:
            with mock.patch("airflow.utils.log.file_task_handler.FileTaskHandler.set_context"):
                self.cloudwatch_task_handler.set_context(self.ti)
                for _ in range(5):
                    self.cloudwatch_task_handler.close()

                mock_log_handler_close.assert_called_once()

    def test_filename_template_for_backward_compatibility(self):
        # filename_template arg support for running the latest provider on airflow 2
        CloudwatchTaskHandler(
            self.local_log_location,
            f"arn:aws:logs:{self.region_name}:11111111:log-group:{self.remote_log_group}",
            filename_template=None,
        )

    def test_event_to_str(self):
        handler = self.cloudwatch_task_handler
        current_time = int(time.time()) * 1000
        events = [
            {"timestamp": current_time - 2000, "message": "First"},
            {"timestamp": current_time - 1000, "message": "Second"},
            {"timestamp": current_time, "message": "Third"},
        ]
        assert [handler._event_to_str(event) for event in events] == (
            [
                f"[{get_time_str(current_time - 2000)}] First",
                f"[{get_time_str(current_time - 1000)}] Second",
                f"[{get_time_str(current_time)}] Third",
            ]
        )


def generate_log_events(conn, log_group_name, log_stream_name, log_events):
    conn.create_log_group(logGroupName=log_group_name)
    conn.create_log_stream(logGroupName=log_group_name, logStreamName=log_stream_name)
    conn.put_log_events(logGroupName=log_group_name, logStreamName=log_stream_name, logEvents=log_events)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.hooks.ssm import SsmHook
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.ssm import SsmRunCommandTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class SsmRunCommandOperator(AwsBaseOperator[SsmHook]):
    """
    Executes the SSM Run Command to perform actions on managed instances.

    .. seealso::
        For more information on how to use this operator, take a look at the
        guide:
        :ref:`howto/operator:SsmRunCommandOperator`

    :param document_name: The name of the Amazon Web Services Systems Manager
        document (SSM document) to run.
    :param run_command_kwargs: Optional parameters to pass to the send_command
        API.

    :param wait_for_completion: Whether to wait for cluster to stop.
        (default: True)
    :param waiter_delay: Time in seconds to wait between status checks.
        (default: 120)
    :param waiter_max_attempts: Maximum number of attempts to check for job
        completion. (default: 75)
    :param fail_on_nonzero_exit: If True (default), the operator will fail when
        the command returns a non-zero exit code. If False, the operator will
        complete successfully regardless of the command exit code, allowing
        downstream tasks to handle exit codes for workflow routing. Note that
        AWS-level failures (Cancelled, TimedOut) will still raise exceptions
        even when this is False. (default: True)
    :param deferrable: If True, the operator will wait asynchronously for the
        cluster to stop. This implies waiting for completion. This mode
        requires aiobotocore module to be installed. (default: False)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used.
        If running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default
        boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore
        client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = SsmHook
    template_fields: Sequence[str] = aws_template_fields(
        "document_name",
        "run_command_kwargs",
    )

    def __init__(
        self,
        *,
        document_name: str,
        run_command_kwargs: dict[str, Any] | None = None,
        wait_for_completion: bool = True,
        waiter_delay: int = 120,
        waiter_max_attempts: int = 75,
        fail_on_nonzero_exit: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.wait_for_completion = wait_for_completion
        self.waiter_delay = waiter_delay
        self.waiter_max_attempts = waiter_max_attempts
        self.fail_on_nonzero_exit = fail_on_nonzero_exit
        self.deferrable = deferrable

        self.document_name = document_name
        self.run_command_kwargs = run_command_kwargs or {}

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        event = validate_execute_complete_event(event)

        if event["status"] != "success":
            raise RuntimeError(f"Error while running run command: {event}")

        self.log.info("SSM run command `%s` completed.", event["command_id"])
        return event["command_id"]

    def execute(self, context: Context):
        response = self.hook.conn.send_command(
            DocumentName=self.document_name,
            **self.run_command_kwargs,
        )

        command_id = response["Command"]["CommandId"]
        task_description = f"SSM run command {command_id} to complete."

        if self.deferrable:
            self.log.info("Deferring for %s", task_description)
            self.defer(
                trigger=SsmRunCommandTrigger(
                    command_id=command_id,
                    waiter_delay=self.waiter_delay,
                    waiter_max_attempts=self.waiter_max_attempts,
                    fail_on_nonzero_exit=self.fail_on_nonzero_exit,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )

        elif self.wait_for_completion:
            self.log.info("Waiting for %s", task_description)
            waiter = self.hook.get_waiter("command_executed")

            instance_ids = response["Command"]["InstanceIds"]
            for instance_id in instance_ids:
                try:
                    waiter.wait(
                        CommandId=command_id,
                        InstanceId=instance_id,
                        WaiterConfig={
                            "Delay": self.waiter_delay,
                            "MaxAttempts": self.waiter_max_attempts,
                        },
                    )
                except WaiterError:
                    if not self.fail_on_nonzero_exit:
                        # Enhanced mode: distinguish between AWS-level and command-level failures
                        invocation = self.hook.get_command_invocation(command_id, instance_id)
                        status = invocation.get("Status", "")

                        # AWS-level failures should always raise
                        if SsmHook.is_aws_level_failure(status):
                            raise

                        # Command-level failure - tolerate it in enhanced mode
                        self.log.info(
                            "Command completed with status %s (exit code: %s). "
                            "Continuing due to fail_on_nonzero_exit=False",
                            status,
                            invocation.get("ResponseCode", "unknown"),
                        )
                    else:
                        # Traditional mode: all failures raise
                        raise

        return command_id


class SsmGetCommandInvocationOperator(AwsBaseOperator[SsmHook]):
    """
    Retrieves the output and execution details of an SSM command invocation.

    .. seealso::
        For more information on how to use this operator, take a look at the
        guide:
        :ref:`howto/operator:SsmGetCommandInvocationOperator`

    :param command_id: The ID of the SSM command to retrieve output for.
    :param instance_id: The ID of the specific instance to retrieve output
        for. If not provided, retrieves output from all instances that
        executed the command.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used.
        If running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default
        boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore
        client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = SsmHook
    template_fields: Sequence[str] = aws_template_fields(
        "command_id",
        "instance_id",
    )

    def __init__(
        self,
        *,
        command_id: str,
        instance_id: str | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.command_id = command_id
        self.instance_id = instance_id

    def execute(self, context: Context) -> dict[str, Any]:
        """Execute the operator to retrieve command invocation output."""
        if self.instance_id:
            self.log.info(
                "Retrieving output for command %s on instance %s",
                self.command_id,
                self.instance_id,
            )
            invocations = [{"InstanceId": self.instance_id}]
        else:
            self.log.info("Retrieving output for command %s from all instances", self.command_id)
            response = self.hook.list_command_invocations(self.command_id)
            invocations = response.get("CommandInvocations", [])

        output_data: dict[str, Any] = {"command_id": self.command_id, "invocations": []}

        for invocation in invocations:
            instance_id = invocation["InstanceId"]
            try:
                invocation_details = self.hook.get_command_invocation(self.command_id, instance_id)
                output_data["invocations"].append(
                    {
                        "instance_id": instance_id,
                        "status": invocation_details.get("Status", ""),
                        "response_code": invocation_details.get("ResponseCode", ""),
                        "standard_output": invocation_details.get("StandardOutputContent", ""),
                        "standard_error": invocation_details.get("StandardErrorContent", ""),
                        "execution_start_time": invocation_details.get("ExecutionStartDateTime", ""),
                        "execution_end_time": invocation_details.get("ExecutionEndDateTime", ""),
                        "document_name": invocation_details.get("DocumentName", ""),
                        "comment": invocation_details.get("Comment", ""),
                    }
                )
            except Exception as e:
                self.log.warning("Failed to get output for instance %s: %s", instance_id, e)
                output_data["invocations"].append({"instance_id": instance_id, "error": str(e)})

        return output_data



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest

from airflow.providers.amazon.aws.triggers.sqs import SqsSensorTrigger

pytest.importorskip("airflow.providers.common.messaging.providers.base_provider")


def test_message_sqs_queue_create():
    from airflow.providers.amazon.aws.queues.sqs import SqsMessageQueueProvider
    from airflow.providers.common.messaging.providers.base_provider import BaseMessageQueueProvider

    provider = SqsMessageQueueProvider()
    assert isinstance(provider, BaseMessageQueueProvider)


def test_message_sqs_queue_matches():
    from airflow.providers.amazon.aws.queues.sqs import SqsMessageQueueProvider

    provider = SqsMessageQueueProvider()
    assert provider.queue_matches("https://sqs.us-east-1.amazonaws.com/123456789012/my-queue")
    assert not provider.queue_matches("https://sqs.us-east-1.amazonaws.com/123456789012")
    assert not provider.queue_matches("https://sqs.us-east-1.amazonaws.com/123456789012/")
    assert not provider.queue_matches("https://sqs.us-east-1.amazonaws.com/")


@pytest.mark.parametrize(
    ("scheme", "expected_result"),
    [
        pytest.param("sqs", True, id="sqs_scheme"),
        pytest.param("kafka", False, id="kafka_scheme"),
        pytest.param("redis+pubsub", False, id="redis_scheme"),
        pytest.param("unknown", False, id="unknown_scheme"),
    ],
)
def test_message_sqs_scheme_matches(scheme, expected_result):
    """Test the scheme_matches method with various schemes."""
    from airflow.providers.amazon.aws.queues.sqs import SqsMessageQueueProvider

    provider = SqsMessageQueueProvider()
    assert provider.scheme_matches(scheme) == expected_result


def test_message_sqs_queue_trigger_class():
    from airflow.providers.amazon.aws.queues.sqs import SqsMessageQueueProvider

    provider = SqsMessageQueueProvider()
    assert provider.trigger_class() == SqsSensorTrigger


def test_message_sqs_queue_trigger_kwargs():
    from airflow.providers.amazon.aws.queues.sqs import SqsMessageQueueProvider

    provider = SqsMessageQueueProvider()
    assert provider.trigger_kwargs("https://sqs.us-east-1.amazonaws.com/123456789012/my-queue") == {
        "sqs_queue": "https://sqs.us-east-1.amazonaws.com/123456789012/my-queue",
    }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.hooks.emr import EmrServerlessHook
from airflow.providers.amazon.aws.operators.emr import (
    EmrServerlessCreateApplicationOperator,
    EmrServerlessDeleteApplicationOperator,
    EmrServerlessStartJobOperator,
    EmrServerlessStopApplicationOperator,
)
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator
from airflow.providers.amazon.aws.sensors.emr import EmrServerlessApplicationSensor, EmrServerlessJobSensor
from airflow.providers.common.compat.sdk import DAG, chain

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_emr_serverless"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"


sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context[ENV_ID_KEY]
    role_arn = test_context[ROLE_ARN_KEY]
    bucket_name = f"{env_id}-emr-serverless-bucket"
    region = boto3.session.Session().region_name
    entryPoint = f"s3://{region}.elasticmapreduce/emr-containers/samples/wordcount/scripts/wordcount.py"
    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=bucket_name)

    SPARK_JOB_DRIVER = {
        "sparkSubmit": {
            "entryPoint": entryPoint,
            "entryPointArguments": [f"s3://{bucket_name}/output"],
            "sparkSubmitParameters": "--conf spark.executor.cores=1 --conf spark.executor.memory=4g\
                --conf spark.driver.cores=1 --conf spark.driver.memory=4g --conf spark.executor.instances=1",
        }
    }

    SPARK_CONFIGURATION_OVERRIDES = {
        "monitoringConfiguration": {"s3MonitoringConfiguration": {"logUri": f"s3://{bucket_name}/logs"}}
    }

    # [START howto_operator_emr_serverless_create_application]
    emr_serverless_app = EmrServerlessCreateApplicationOperator(
        task_id="create_emr_serverless_task",
        release_label="emr-6.6.0",
        job_type="SPARK",
        config={"name": "new_application"},
    )
    # [END howto_operator_emr_serverless_create_application]

    # EmrServerlessCreateApplicationOperator waits by default, setting as False to test the Sensor below.
    emr_serverless_app.wait_for_completion = False

    emr_serverless_app_id = emr_serverless_app.output

    # [START howto_sensor_emr_serverless_application]
    wait_for_app_creation = EmrServerlessApplicationSensor(
        task_id="wait_for_app_creation",
        application_id=emr_serverless_app_id,
    )
    # [END howto_sensor_emr_serverless_application]
    wait_for_app_creation.poke_interval = 1

    # [START howto_operator_emr_serverless_start_job]
    start_job = EmrServerlessStartJobOperator(
        task_id="start_emr_serverless_job",
        application_id=emr_serverless_app_id,
        execution_role_arn=role_arn,
        job_driver=SPARK_JOB_DRIVER,
        configuration_overrides=SPARK_CONFIGURATION_OVERRIDES,
    )
    # [END howto_operator_emr_serverless_start_job]
    start_job.wait_for_completion = False

    # [START howto_sensor_emr_serverless_job]
    wait_for_job = EmrServerlessJobSensor(
        task_id="wait_for_job",
        application_id=emr_serverless_app_id,
        job_run_id=start_job.output,
        # the default is to wait for job completion, here we just wait for the job to be running.
        target_states={*EmrServerlessHook.JOB_SUCCESS_STATES, "RUNNING"},
    )
    # [END howto_sensor_emr_serverless_job]
    wait_for_job.poke_interval = 10

    # [START howto_operator_emr_serverless_stop_application]
    stop_app = EmrServerlessStopApplicationOperator(
        task_id="stop_application",
        application_id=emr_serverless_app_id,
        force_stop=True,
    )
    # [END howto_operator_emr_serverless_stop_application]
    stop_app.waiter_delay = 1

    # [START howto_operator_emr_serverless_delete_application]
    delete_app = EmrServerlessDeleteApplicationOperator(
        task_id="delete_application",
        application_id=emr_serverless_app_id,
    )
    # [END howto_operator_emr_serverless_delete_application]
    delete_app.waiter_delay = 1
    delete_app.trigger_rule = TriggerRule.ALL_DONE

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=bucket_name,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket,
        # TEST BODY
        emr_serverless_app,
        wait_for_app_creation,
        start_job,
        wait_for_job,
        stop_app,
        # TEST TEARDOWN
        delete_app,
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any
from urllib.parse import urlparse

from airflow.providers.amazon.aws.hooks.athena import AthenaHook
from airflow.providers.amazon.aws.links.athena import AthenaQueryResultsLink
from airflow.providers.amazon.aws.operators.base_aws import AwsBaseOperator
from airflow.providers.amazon.aws.triggers.athena import AthenaTrigger
from airflow.providers.amazon.aws.utils import validate_execute_complete_event
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.providers.common.compat.openlineage.facet import BaseFacet, Dataset, DatasetFacet
    from airflow.providers.openlineage.extractors.base import OperatorLineage
    from airflow.sdk import Context


class AthenaOperator(AwsBaseOperator[AthenaHook]):
    """
    An operator that submits a Trino/Presto query to Amazon Athena.

    .. note:: if the task is killed while it runs, it'll cancel the athena query that was launched,
        EXCEPT if running in deferrable mode.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:AthenaOperator`

    :param query: Trino/Presto query to be run on Amazon Athena. (templated)
    :param database: Database to select. (templated)
    :param catalog: Catalog to select. (templated)
    :param output_location: s3 path to write the query results into. (templated)
        To run the query, you must specify the query results location using one of the ways:
        either for individual queries using either this setting (client-side),
        or in the workgroup, using WorkGroupConfiguration.
        If none of them is set, Athena issues an error that no output location is provided
    :param client_request_token: Unique token created by user to avoid multiple executions of same query
    :param workgroup: Athena workgroup in which query will be run. (templated)
    :param query_execution_context: Context in which query need to be run
    :param result_configuration: Dict with path to store results in and config related to encryption
    :param sleep_time: Time (in seconds) to wait between two consecutive calls to check query status on Athena
    :param max_polling_attempts: Number of times to poll for query state before function exits
        To limit task execution time, use execution_timeout.
    :param log_query: Whether to log athena query and other execution params when it's executed.
        Defaults to *True*.
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    aws_hook_class = AthenaHook
    ui_color = "#44b5e2"
    template_fields: Sequence[str] = aws_template_fields(
        "query", "database", "output_location", "workgroup", "catalog"
    )
    template_ext: Sequence[str] = (".sql",)
    template_fields_renderers = {"query": "sql"}
    operator_extra_links = (AthenaQueryResultsLink(),)

    def __init__(
        self,
        *,
        query: str,
        database: str,
        output_location: str | None = None,
        client_request_token: str | None = None,
        workgroup: str = "primary",
        query_execution_context: dict[str, str] | None = None,
        result_configuration: dict[str, Any] | None = None,
        sleep_time: int = 30,
        max_polling_attempts: int | None = None,
        log_query: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        catalog: str = "AwsDataCatalog",
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.query = query
        self.database = database
        self.output_location = output_location
        self.client_request_token = client_request_token
        self.workgroup = workgroup
        self.query_execution_context = query_execution_context or {}
        self.result_configuration = result_configuration or {}
        self.sleep_time = sleep_time
        self.max_polling_attempts = max_polling_attempts or 999999
        self.query_execution_id: str | None = None
        self.log_query: bool = log_query
        self.deferrable = deferrable
        self.catalog: str = catalog

    @property
    def _hook_parameters(self) -> dict[str, Any]:
        return {**super()._hook_parameters, "log_query": self.log_query}

    def execute(self, context: Context) -> str | None:
        """Run Trino/Presto Query on Amazon Athena."""
        self.query_execution_context["Database"] = self.database
        self.query_execution_context["Catalog"] = self.catalog
        if self.output_location:
            self.result_configuration["OutputLocation"] = self.output_location
        self.query_execution_id = self.hook.run_query(
            self.query,
            self.query_execution_context,
            self.result_configuration,
            self.client_request_token,
            self.workgroup,
        )
        AthenaQueryResultsLink.persist(
            context=context,
            operator=self,
            region_name=self.hook.conn_region_name,
            aws_partition=self.hook.conn_partition,
            query_execution_id=self.query_execution_id,
        )

        if self.deferrable:
            self.defer(
                trigger=AthenaTrigger(
                    query_execution_id=self.query_execution_id,
                    waiter_delay=self.sleep_time,
                    waiter_max_attempts=self.max_polling_attempts,
                    aws_conn_id=self.aws_conn_id,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="execute_complete",
            )
        # implicit else:
        query_status = self.hook.poll_query_status(
            self.query_execution_id,
            max_polling_attempts=self.max_polling_attempts,
            sleep_time=self.sleep_time,
        )

        if query_status in AthenaHook.FAILURE_STATES:
            error_message = self.hook.get_state_change_reason(self.query_execution_id)
            raise AirflowException(
                f"Final state of Athena job is {query_status}, query_execution_id is "
                f"{self.query_execution_id}. Error: {error_message}"
            )
        if not query_status or query_status in AthenaHook.INTERMEDIATE_STATES:
            raise AirflowException(
                f"Final state of Athena job is {query_status}. Max tries of poll status exceeded, "
                f"query_execution_id is {self.query_execution_id}."
            )

        return self.query_execution_id

    def execute_complete(self, context: Context, event: dict[str, Any] | None = None) -> str:
        validated_event = validate_execute_complete_event(event)

        if validated_event["status"] != "success":
            raise AirflowException(
                f"Error while waiting for operation on cluster to complete: {validated_event}"
            )

        # Save query_execution_id to be later used by listeners
        self.query_execution_id = validated_event["value"]
        return validated_event["value"]

    def on_kill(self) -> None:
        """Cancel the submitted Amazon Athena query."""
        if self.query_execution_id:
            self.log.info("Received a kill signal.")
            response = self.hook.stop_query(self.query_execution_id)
            http_status_code = None
            try:
                http_status_code = response["ResponseMetadata"]["HTTPStatusCode"]
            except Exception:
                self.log.exception(
                    "Exception while cancelling query. Query execution id: %s", self.query_execution_id
                )
            finally:
                if http_status_code is None or http_status_code != 200:
                    self.log.error("Unable to request query cancel on athena. Exiting")
                else:
                    self.log.info(
                        "Polling Athena for query with id %s to reach final state", self.query_execution_id
                    )
                    self.hook.poll_query_status(self.query_execution_id, sleep_time=self.sleep_time)

    def get_openlineage_facets_on_complete(self, _) -> OperatorLineage:
        """
        Retrieve OpenLineage data by parsing SQL queries and enriching them with Athena API.

        In addition to CTAS query, query and calculation results are stored in S3 location.
        For that reason additional output is attached with this location. Instead of using the complete
        path where the results are saved (user's prefix + some UUID), we are creating a dataset with the
        user-provided path only. This should make it easier to match this dataset across different processes.
        """
        from airflow.providers.common.compat.openlineage.facet import (
            Dataset,
            Error,
            ExternalQueryRunFacet,
            ExtractionErrorRunFacet,
            SQLJobFacet,
        )
        from airflow.providers.openlineage.extractors.base import OperatorLineage
        from airflow.providers.openlineage.sqlparser import SQLParser

        sql_parser = SQLParser(dialect="generic")

        job_facets: dict[str, BaseFacet] = {"sql": SQLJobFacet(query=sql_parser.normalize_sql(self.query))}
        parse_result = sql_parser.parse(sql=self.query)

        if not parse_result:
            return OperatorLineage(job_facets=job_facets)

        run_facets: dict[str, BaseFacet] = {}
        if parse_result.errors:
            run_facets["extractionError"] = ExtractionErrorRunFacet(
                totalTasks=1,
                failedTasks=len(parse_result.errors),
                errors=[
                    Error(
                        errorMessage=error.message,
                        stackTrace=None,
                        task=error.origin_statement,
                        taskNumber=error.index,
                    )
                    for error in parse_result.errors
                ],
            )

        inputs: list[Dataset] = list(
            filter(
                None,
                [
                    self.get_openlineage_dataset(table.schema or self.database, table.name)
                    for table in parse_result.in_tables
                ],
            )
        )

        outputs: list[Dataset] = list(
            filter(
                None,
                [
                    self.get_openlineage_dataset(table.schema or self.database, table.name)
                    for table in parse_result.out_tables
                ],
            )
        )

        if self.query_execution_id:
            run_facets["externalQuery"] = ExternalQueryRunFacet(
                externalQueryId=self.query_execution_id, source="awsathena"
            )

        if self.output_location:
            parsed = urlparse(self.output_location)
            outputs.append(Dataset(namespace=f"{parsed.scheme}://{parsed.netloc}", name=parsed.path or "/"))

        return OperatorLineage(job_facets=job_facets, run_facets=run_facets, inputs=inputs, outputs=outputs)

    def get_openlineage_dataset(self, database, table) -> Dataset | None:
        from airflow.providers.common.compat.openlineage.facet import (
            Dataset,
            Identifier,
            SchemaDatasetFacet,
            SchemaDatasetFacetFields,
            SymlinksDatasetFacet,
        )

        client = self.hook.get_conn()
        try:
            table_metadata = client.get_table_metadata(
                CatalogName=self.catalog, DatabaseName=database, TableName=table
            )

            # Dataset has also its' physical location which we can add in symlink facet.
            s3_location = table_metadata["TableMetadata"]["Parameters"]["location"]
            parsed_path = urlparse(s3_location)
            facets: dict[str, DatasetFacet] = {
                "symlinks": SymlinksDatasetFacet(
                    identifiers=[
                        Identifier(
                            namespace=f"{parsed_path.scheme}://{parsed_path.netloc}",
                            name=str(parsed_path.path),
                            type="TABLE",
                        )
                    ]
                )
            }
            fields = [
                SchemaDatasetFacetFields(
                    name=column["Name"],
                    type=column["Type"],
                    description=column.get("Comment"),
                )
                for column in table_metadata["TableMetadata"]["Columns"]
            ]
            if fields:
                facets["schema"] = SchemaDatasetFacet(fields=fields)
            return Dataset(
                namespace=f"awsathena://athena.{self.hook.region_name}.amazonaws.com",
                name=".".join(filter(None, (self.catalog, database, table))),
                facets=facets,
            )

        except Exception as e:
            self.log.error("Cannot retrieve table metadata from Athena.Client. %s", e)
            return None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from unittest import mock

import boto3
import pytest

from airflow.providers.amazon.aws.hooks.glue_databrew import GlueDataBrewHook

RUNNING_STATES = ["STARTING", "RUNNING", "STOPPING"]
TERMINAL_STATES = ["STOPPED", "SUCCEEDED", "FAILED"]


class TestCustomDataBrewWaiters:
    """Test waiters from ``amazon/aws/waiters/glue.json``."""

    JOB_NAME = "test_job"
    RUN_ID = "123"

    @pytest.fixture(autouse=True)
    def _setup_test_cases(self, monkeypatch):
        self.client = boto3.client("databrew", region_name="eu-west-3")
        monkeypatch.setattr(GlueDataBrewHook, "conn", self.client)

    def test_service_waiters(self):
        hook_waiters = GlueDataBrewHook(aws_conn_id=None).list_waiters()
        assert "job_complete" in hook_waiters

    @pytest.fixture
    def mock_describe_job_runs(self):
        """Mock ``GlueDataBrewHook.Client.describe_job_run`` method."""
        with mock.patch.object(self.client, "describe_job_run") as m:
            yield m

    @staticmethod
    def describe_jobs(status: str):
        """
        Helper function for generate minimal DescribeJobRun response for a single job.

        https://docs.aws.amazon.com/databrew/latest/dg/API_DescribeJobRun.html
        """
        return {"State": status}

    def test_job_succeeded(self, mock_describe_job_runs):
        """Test job succeeded"""
        mock_describe_job_runs.side_effect = [
            self.describe_jobs(RUNNING_STATES[1]),
            self.describe_jobs(TERMINAL_STATES[1]),
        ]
        waiter = GlueDataBrewHook(aws_conn_id=None).get_waiter("job_complete")
        waiter.wait(name=self.JOB_NAME, runId=self.RUN_ID, WaiterConfig={"Delay": 0.2, "MaxAttempts": 2})



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.hooks.emr import EmrContainerHook
from airflow.providers.amazon.aws.sensors.emr import EmrContainerSensor
from airflow.providers.amazon.aws.triggers.emr import EmrContainerTrigger
from airflow.providers.common.compat.sdk import AirflowException, TaskDeferred


class TestEmrContainerSensor:
    def setup_method(self):
        self.sensor = EmrContainerSensor(
            task_id="test_emrcontainer_sensor",
            virtual_cluster_id="vzwemreks",
            job_id="job1234",
            poll_interval=5,
            max_retries=1,
            aws_conn_id="aws_default",
        )
        # We're mocking all actual AWS calls and don't need a connection. This
        # avoids an Airflow warning about connection cannot be found.
        self.sensor.hook.get_connection = lambda _: None

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("PENDING",))
    def test_poke_pending(self, mock_check_query_status):
        assert not self.sensor.poke(None)

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("SUBMITTED",))
    def test_poke_submitted(self, mock_check_query_status):
        assert not self.sensor.poke(None)

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("RUNNING",))
    def test_poke_running(self, mock_check_query_status):
        assert not self.sensor.poke(None)

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("COMPLETED",))
    def test_poke_completed(self, mock_check_query_status):
        assert self.sensor.poke(None)

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("FAILED",))
    def test_poke_failed(self, mock_check_query_status):
        with pytest.raises(AirflowException) as ctx:
            self.sensor.poke(None)
        assert "EMR Containers sensor failed" in str(ctx.value)
        assert "FAILED" in str(ctx.value)

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("CANCELLED",))
    def test_poke_cancelled(self, mock_check_query_status):
        with pytest.raises(AirflowException) as ctx:
            self.sensor.poke(None)
        assert "EMR Containers sensor failed" in str(ctx.value)
        assert "CANCELLED" in str(ctx.value)

    @mock.patch.object(EmrContainerHook, "check_query_status", side_effect=("CANCEL_PENDING",))
    def test_poke_cancel_pending(self, mock_check_query_status):
        with pytest.raises(AirflowException) as ctx:
            self.sensor.poke(None)
        assert "EMR Containers sensor failed" in str(ctx.value)
        assert "CANCEL_PENDING" in str(ctx.value)

    @mock.patch("airflow.providers.amazon.aws.sensors.emr.EmrContainerSensor.poke")
    def test_sensor_defer(self, mock_poke):
        self.sensor.deferrable = True
        mock_poke.return_value = False
        with pytest.raises(TaskDeferred) as e:
            self.sensor.execute(context=None)
        assert isinstance(e.value.trigger, EmrContainerTrigger), (
            f"{e.value.trigger} is not a EmrContainerTrigger"
        )

    @mock.patch("airflow.providers.amazon.aws.sensors.emr.EmrContainerSensor.poke")
    def test_sensor_defer_with_timeout(self, mock_poke):
        self.sensor.deferrable = True
        mock_poke.return_value = False
        self.sensor.max_retries = 1000

        with pytest.raises(TaskDeferred) as e:
            self.sensor.execute(context=None)

        trigger = e.value.trigger
        assert isinstance(trigger, EmrContainerTrigger), f"{trigger} is not a EmrContainerTrigger"
        assert trigger.waiter_delay == self.sensor.poll_interval
        assert trigger.attempts == self.sensor.max_retries



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
import re
from datetime import datetime, timezone
from enum import Enum
from importlib import metadata
from typing import Any

from airflow.providers.common.compat.sdk import AirflowException
from airflow.utils.helpers import prune_dict
from airflow.version import version

log = logging.getLogger(__name__)


def trim_none_values(obj: dict):
    return prune_dict(obj)


def datetime_to_epoch(date_time: datetime) -> int:
    """Convert a datetime object to an epoch integer (seconds)."""
    return int(date_time.timestamp())


def datetime_to_epoch_ms(date_time: datetime) -> int:
    """Convert a datetime object to an epoch integer (milliseconds)."""
    return int(date_time.timestamp() * 1_000)


def datetime_to_epoch_utc_ms(date_time: datetime) -> int:
    """Convert a datetime object to an epoch integer (milliseconds) in UTC timezone."""
    return int(date_time.replace(tzinfo=timezone.utc).timestamp() * 1_000)


def datetime_to_epoch_us(date_time: datetime) -> int:
    """Convert a datetime object to an epoch integer (microseconds)."""
    return int(date_time.timestamp() * 1_000_000)


def get_airflow_version() -> tuple[int, ...]:
    match = re.match(r"(\d+)\.(\d+)\.(\d+)", version)
    if match is None:  # Not theoratically possible.
        raise RuntimeError(f"Broken Airflow version: {version}")
    return tuple(int(x) for x in match.groups())


def get_botocore_version() -> tuple[int, ...]:
    """Return the version number of the installed botocore package in the form of a tuple[int,...]."""
    return tuple(map(int, metadata.version("botocore").split(".")[:3]))


def validate_execute_complete_event(event: dict[str, Any] | None = None) -> dict[str, Any]:
    if event is None:
        err_msg = "Trigger error: event is None"
        log.error(err_msg)
        raise AirflowException(err_msg)
    return event


class _StringCompareEnum(Enum):
    """
    An Enum class which can be compared with regular `str` and subclasses.

    This class avoids multiple inheritance such as AwesomeEnum(str, Enum)
    which does not work well with templated_fields and Jinja templates.
    """

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        return super().__eq__(other)

    def __hash__(self):
        return super().__hash__()  # Need to set because we redefine __eq__



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock
from unittest.mock import AsyncMock

import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.hooks.ecs import EcsHook
from airflow.providers.amazon.aws.hooks.logs import AwsLogsHook
from airflow.providers.amazon.aws.triggers.ecs import (
    TaskDoneTrigger,
)
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from airflow.triggers.base import TriggerEvent


class TestTaskDoneTrigger:
    @pytest.mark.asyncio
    @mock.patch.object(EcsHook, "get_async_conn")
    # this mock is only necessary to avoid a "No module named 'aiobotocore'" error in the LatestBoto CI step
    @mock.patch.object(AwsLogsHook, "get_async_conn")
    async def test_run_until_error(self, _, client_mock):
        a_mock = mock.MagicMock()
        client_mock.return_value.__aenter__.return_value = a_mock
        wait_mock = AsyncMock()
        wait_mock.side_effect = [
            WaiterError("name", "reason", {"tasks": [{"lastStatus": "my_status"}]}),
            WaiterError("name", "reason", {"tasks": [{"lastStatus": "my_status"}]}),
            WaiterError("terminal failure", "reason", {}),
        ]
        a_mock.get_waiter().wait = wait_mock

        trigger = TaskDoneTrigger("cluster", "task_arn", 0, 10, None, None)
        generator = trigger.run()
        with pytest.raises(WaiterError):
            await generator.asend(None)
        assert wait_mock.call_count == 3

    @pytest.mark.asyncio
    @mock.patch.object(EcsHook, "get_async_conn")
    # this mock is only necessary to avoid a "No module named 'aiobotocore'" error in the LatestBoto CI step
    @mock.patch.object(AwsLogsHook, "get_async_conn")
    async def test_run_until_timeout(self, _, client_mock):
        a_mock = mock.MagicMock()
        client_mock.return_value.__aenter__.return_value = a_mock
        wait_mock = AsyncMock()
        wait_mock.side_effect = WaiterError("name", "reason", {"tasks": [{"lastStatus": "my_status"}]})
        a_mock.get_waiter().wait = wait_mock

        trigger = TaskDoneTrigger("cluster", "task_arn", 0, 10, None, None)
        generator = trigger.run()
        with pytest.raises(AirflowException) as err:
            await generator.asend(None)

        assert wait_mock.call_count == 10
        assert "max attempts" in str(err.value)

    @pytest.mark.asyncio
    @mock.patch.object(EcsHook, "get_async_conn")
    # this mock is only necessary to avoid a "No module named 'aiobotocore'" error in the LatestBoto CI step
    @mock.patch.object(AwsLogsHook, "get_async_conn")
    async def test_run_success(self, _, client_mock):
        a_mock = mock.MagicMock()
        client_mock.return_value.__aenter__.return_value = a_mock
        wait_mock = AsyncMock()
        a_mock.get_waiter().wait = wait_mock

        trigger = TaskDoneTrigger("cluster", "my_task_arn", 0, 10, None, None)

        generator = trigger.run()
        response: TriggerEvent = await generator.asend(None)

        assert response.payload["status"] == "success"
        assert response.payload["task_arn"] == "my_task_arn"
        assert response.payload["cluster"] == "cluster"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import zipfile
from datetime import datetime
from io import BytesIO

import boto3

from airflow.providers.amazon.aws.operators.lambda_function import (
    LambdaCreateFunctionOperator,
    LambdaInvokeFunctionOperator,
)
from airflow.providers.amazon.aws.sensors.lambda_function import LambdaFunctionStateSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder, prune_logs

DAG_ID = "example_lambda"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

CODE_CONTENT = """
def test(*args):
    print('Hello')
"""


# Create a zip file containing one file "lambda_function.py" to deploy to the lambda function
def create_zip(content: str):
    with BytesIO() as zip_output:
        with zipfile.ZipFile(zip_output, "w", zipfile.ZIP_DEFLATED) as zip_file:
            info = zipfile.ZipInfo("lambda_function.py")
            info.external_attr = 0o777 << 16
            zip_file.writestr(info, content)
        zip_output.seek(0)
        return zip_output.read()


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_lambda(function_name: str):
    client = boto3.client("lambda")
    client.delete_function(
        FunctionName=function_name,
    )


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    lambda_function_name: str = f"{test_context[ENV_ID_KEY]}-function"
    role_arn = test_context[ROLE_ARN_KEY]

    # [START howto_operator_create_lambda_function]
    create_lambda_function = LambdaCreateFunctionOperator(
        task_id="create_lambda_function",
        function_name=lambda_function_name,
        runtime="python3.9",
        role=role_arn,
        handler="lambda_function.test",
        code={
            "ZipFile": create_zip(CODE_CONTENT),
        },
    )
    # [END howto_operator_create_lambda_function]

    # [START howto_sensor_lambda_function_state]
    wait_lambda_function_state = LambdaFunctionStateSensor(
        task_id="wait_lambda_function_state",
        function_name=lambda_function_name,
    )
    # [END howto_sensor_lambda_function_state]
    wait_lambda_function_state.poke_interval = 1

    # [START howto_operator_invoke_lambda_function]
    invoke_lambda_function = LambdaInvokeFunctionOperator(
        task_id="invoke_lambda_function",
        function_name=lambda_function_name,
        payload=json.dumps({"SampleEvent": {"SampleData": {"Name": "XYZ", "DoB": "1993-01-01"}}}),
    )
    # [END howto_operator_invoke_lambda_function]

    log_cleanup = prune_logs(
        [
            # Format: ('log group name', 'log stream prefix')
            (f"/aws/lambda/{lambda_function_name}", None),
        ],
        force_delete=True,
        retry=True,
    )

    chain(
        # TEST SETUP
        test_context,
        # TEST BODY
        create_lambda_function,
        wait_lambda_function_state,
        invoke_lambda_function,
        # TEST TEARDOWN
        delete_lambda(lambda_function_name),
        log_cleanup,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from unittest import mock

import pytest
from moto import mock_aws

from airflow.configuration import initialize_secrets_backends
from airflow.providers.amazon.aws.secrets.systems_manager import SystemsManagerParameterStoreBackend

from tests_common.test_utils.config import conf_vars

URI_CONNECTION = pytest.param(
    "postgres://my-login:my-pass@my-host:5432/my-schema?param1=val1&param2=val2", id="uri-connection"
)
JSON_CONNECTION = pytest.param(
    json.dumps(
        {
            "conn_type": "postgres",
            "login": "my-login",
            "password": "my-pass",
            "host": "my-host",
            "port": 5432,
            "schema": "my-schema",
            "extra": {"param1": "val1", "param2": "val2"},
        }
    ),
    id="json-connection",
)


class TestSsmSecrets:
    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager."
        "SystemsManagerParameterStoreBackend.get_conn_value"
    )
    def test_aws_ssm_get_connection(self, mock_get_value):
        mock_get_value.return_value = "scheme://user:pass@host:100"
        conn = SystemsManagerParameterStoreBackend().get_connection("fake_conn")
        assert conn.host == "host"

    @mock_aws
    @pytest.mark.parametrize("ssm_value", [JSON_CONNECTION, URI_CONNECTION])
    def test_get_conn_value(self, ssm_value):
        param = {
            "Name": "/airflow/connections/test_postgres",
            "Type": "String",
            "Value": ssm_value,
        }

        ssm_backend = SystemsManagerParameterStoreBackend()
        ssm_backend.client.put_parameter(**param)

        returned_conn_value = ssm_backend.get_conn_value(conn_id="test_postgres")
        assert ssm_value == returned_conn_value

        test_conn = ssm_backend.get_connection(conn_id="test_postgres")
        assert test_conn.conn_id == "test_postgres"
        assert test_conn.conn_type == "postgres"
        assert test_conn.login == "my-login"
        assert test_conn.password == "my-pass"
        assert test_conn.host == "my-host"
        assert test_conn.port == 5432
        assert test_conn.schema == "my-schema"
        assert test_conn.extra_dejson == {"param1": "val1", "param2": "val2"}

    @mock_aws
    def test_get_conn_value_non_existent_key(self):
        """
        Test that if the key with connection ID is not present in SSM,
        SystemsManagerParameterStoreBackend.get_connection should return None
        """
        conn_id = "test_mysql"
        param = {
            "Name": "/airflow/connections/test_postgres",
            "Type": "String",
            "Value": "postgresql://airflow:airflow@host:5432/airflow",
        }

        ssm_backend = SystemsManagerParameterStoreBackend()
        ssm_backend.client.put_parameter(**param)

        assert ssm_backend.get_conn_value(conn_id=conn_id) is None
        assert ssm_backend.get_connection(conn_id=conn_id) is None

    @mock_aws
    def test_get_variable(self):
        param = {"Name": "/airflow/variables/hello", "Type": "String", "Value": "world"}

        ssm_backend = SystemsManagerParameterStoreBackend()
        ssm_backend.client.put_parameter(**param)

        returned_uri = ssm_backend.get_variable("hello")
        assert returned_uri == "world"

    @mock_aws
    def test_get_config(self):
        param = {
            "Name": "/airflow/config/sql_alchemy_conn",
            "Type": "String",
            "Value": "sqlite:///Users/test_user/airflow.db",
        }

        ssm_backend = SystemsManagerParameterStoreBackend()
        ssm_backend.client.put_parameter(**param)

        returned_uri = ssm_backend.get_config("sql_alchemy_conn")
        assert returned_uri == "sqlite:///Users/test_user/airflow.db"

    @mock_aws
    def test_get_variable_secret_string(self):
        param = {"Name": "/airflow/variables/hello", "Type": "SecureString", "Value": "world"}
        ssm_backend = SystemsManagerParameterStoreBackend()
        ssm_backend.client.put_parameter(**param)
        returned_uri = ssm_backend.get_variable("hello")
        assert returned_uri == "world"

    @mock_aws
    def test_get_variable_non_existent_key(self):
        """
        Test that if Variable key is not present in SSM,
        SystemsManagerParameterStoreBackend.get_variables should return None
        """
        param = {"Name": "/airflow/variables/hello", "Type": "String", "Value": "world"}

        ssm_backend = SystemsManagerParameterStoreBackend()
        ssm_backend.client.put_parameter(**param)

        assert ssm_backend.get_variable("test_mysql") is None

    @conf_vars(
        {
            ("secrets", "backend"): "airflow.providers.amazon.aws.secrets.systems_manager."
            "SystemsManagerParameterStoreBackend",
            (
                "secrets",
                "backend_kwargs",
            ): '{"use_ssl": false, "role_arn": "arn:aws:iam::222222222222:role/awesome-role"}',
        }
    )
    @mock.patch("airflow.providers.amazon.aws.hooks.base_aws.SessionFactory")
    def test_passing_client_kwargs(self, mock_session_factory):
        backends = initialize_secrets_backends()
        systems_manager = next(
            backend
            for backend in backends
            if backend.__class__.__name__ == "SystemsManagerParameterStoreBackend"
        )

        # Mock SessionFactory, session and client
        mock_session_factory_instance = mock_session_factory.return_value
        mock_ssm_client = mock.MagicMock(return_value="mock-ssm-client")
        mock_session = mock.MagicMock()
        mock_session.client = mock_ssm_client
        mock_create_session = mock.MagicMock(return_value=mock_session)
        mock_session_factory_instance.create_session = mock_create_session

        systems_manager.client
        assert mock_session_factory.call_count == 1
        mock_session_factory_call_kwargs = mock_session_factory.call_args.kwargs
        assert "conn" in mock_session_factory_call_kwargs
        conn_wrapper = mock_session_factory_call_kwargs["conn"]

        assert conn_wrapper.conn_id == "SystemsManagerParameterStoreBackend__connection"
        assert conn_wrapper.role_arn == "arn:aws:iam::222222222222:role/awesome-role"

        mock_ssm_client.assert_called_once_with(service_name="ssm", use_ssl=False)

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend._get_secret"
    )
    def test_connection_prefix_none_value(self, mock_get_secret):
        """
        Test that if Variable key is not present in SSM,
        SystemsManagerParameterStoreBackend.get_conn_value should return None,
        SystemsManagerParameterStoreBackend._get_secret should not be called
        """
        kwargs = {"connections_prefix": None}

        ssm_backend = SystemsManagerParameterStoreBackend(**kwargs)

        assert ssm_backend.get_conn_value("test_mysql") is None
        mock_get_secret.assert_not_called()

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend._get_secret"
    )
    def test_variable_prefix_none_value(self, mock_get_secret):
        """
        Test that if Variable key is not present in SSM,
        SystemsManagerParameterStoreBackend.get_variables should return None,
        SystemsManagerParameterStoreBackend._get_secret should not be called
        """
        kwargs = {"variables_prefix": None}

        ssm_backend = SystemsManagerParameterStoreBackend(**kwargs)

        assert ssm_backend.get_variable("hello") is None
        mock_get_secret.assert_not_called()

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend._get_secret"
    )
    def test_config_prefix_none_value(self, mock_get_secret):
        """
        Test that if Variable key is not present in SSM,
        SystemsManagerParameterStoreBackend.get_config should return None,
        SystemsManagerParameterStoreBackend._get_secret should not be called
        """
        kwargs = {"config_prefix": None}

        ssm_backend = SystemsManagerParameterStoreBackend(**kwargs)

        assert ssm_backend.get_config("config") is None
        mock_get_secret.assert_not_called()

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend.client",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        ("connection_id", "connections_lookup_pattern", "num_client_calls"),
        [
            ("test", "test", 1),
            ("test", ".*", 1),
            ("test", "T.*", 1),
            ("test", "dummy-pattern", 0),
            ("test", None, 1),
        ],
    )
    def test_connection_lookup_pattern(
        self, mock_client, connection_id, connections_lookup_pattern, num_client_calls
    ):
        """
        Test that if Connection ID is looked up in AWS Parameter Store
        """
        mock_client().get_parameter.return_value = {
            "Parameter": {
                "Value": None,
            },
        }
        kwargs = {"connections_lookup_pattern": connections_lookup_pattern}

        secrets_manager_backend = SystemsManagerParameterStoreBackend(**kwargs)
        secrets_manager_backend.get_conn_value(connection_id)
        assert mock_client().get_parameter.call_count == num_client_calls

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend.client",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        ("variable_key", "variables_lookup_pattern", "num_client_calls"),
        [
            ("test", "test", 1),
            ("test", ".*", 1),
            ("test", "T.*", 1),
            ("test", "dummy-pattern", 0),
            ("test", None, 1),
        ],
    )
    def test_variable_lookup_pattern(
        self, mock_client, variable_key, variables_lookup_pattern, num_client_calls
    ):
        """
        Test that if Variable key is looked up in AWS Parameter Store
        """
        mock_client().get_parameter.return_value = {
            "Parameter": {
                "Value": None,
            },
        }
        kwargs = {"variables_lookup_pattern": variables_lookup_pattern}

        secrets_manager_backend = SystemsManagerParameterStoreBackend(**kwargs)
        secrets_manager_backend.get_variable(variable_key)
        assert mock_client().get_parameter.call_count == num_client_calls

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend.client",
        new_callable=mock.PropertyMock,
    )
    @pytest.mark.parametrize(
        ("config_key", "config_lookup_pattern", "num_client_calls"),
        [
            ("test", "test", 1),
            ("test", ".*", 1),
            ("test", "T.*", 1),
            ("test", "dummy-pattern", 0),
            ("test", None, 1),
        ],
    )
    def test_config_lookup_pattern(self, mock_client, config_key, config_lookup_pattern, num_client_calls):
        """
        Test that if Variable key is looked up in AWS Parameter Store
        """
        mock_client().get_parameter.return_value = {
            "Parameter": {
                "Value": None,
            },
        }
        kwargs = {"config_lookup_pattern": config_lookup_pattern}

        secrets_manager_backend = SystemsManagerParameterStoreBackend(**kwargs)
        secrets_manager_backend.get_config(config_key)
        assert mock_client().get_parameter.call_count == num_client_calls

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend.client",
        new_callable=mock.PropertyMock,
    )
    def test_connection_prefix_with_no_leading_slash(self, mock_client):
        """
        Test that if Connection ID is looked up in AWS Parameter Store with the added leading "/"
        """
        mock_client().get_parameter.return_value = {
            "Parameter": {
                "Value": None,
            },
        }
        kwargs = {"connections_prefix": "airflow/connections"}

        secrets_manager_backend = SystemsManagerParameterStoreBackend(**kwargs)
        secrets_manager_backend.get_conn_value("test_mysql")
        mock_client().get_parameter.assert_called_with(
            Name="/airflow/connections/test_mysql", WithDecryption=True
        )

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend.client",
        new_callable=mock.PropertyMock,
    )
    def test_variable_prefix_with_no_leading_slash(self, mock_client):
        """
        Test that if Variable key is looked up in AWS Parameter Store with the added leading "/"
        """
        mock_client().get_parameter.return_value = {
            "Parameter": {
                "Value": None,
            },
        }
        kwargs = {"variables_prefix": "airflow/variables"}

        secrets_manager_backend = SystemsManagerParameterStoreBackend(**kwargs)
        secrets_manager_backend.get_variable("hello")
        mock_client().get_parameter.assert_called_with(Name="/airflow/variables/hello", WithDecryption=True)

    @mock.patch(
        "airflow.providers.amazon.aws.secrets.systems_manager.SystemsManagerParameterStoreBackend.client",
        new_callable=mock.PropertyMock,
    )
    def test_config_prefix_with_no_leading_slash(self, mock_client):
        """
        Test that if Config key is looked up in AWS Parameter Store with the added leading "/"
        """
        mock_client().get_parameter.return_value = {
            "Parameter": {
                "Value": None,
            },
        }
        kwargs = {"config_prefix": "airflow/config"}

        secrets_manager_backend = SystemsManagerParameterStoreBackend(**kwargs)
        secrets_manager_backend.get_config("config")
        mock_client().get_parameter.assert_called_with(Name="/airflow/config/config", WithDecryption=True)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import logging
from typing import Any

import anyio
from fastapi import HTTPException, Request, status
from fastapi.responses import RedirectResponse

from airflow.api_fastapi.app import (
    AUTH_MANAGER_FASTAPI_APP_PREFIX,
    get_auth_manager,
)
from airflow.api_fastapi.auth.managers.base_auth_manager import COOKIE_NAME_JWT_TOKEN
from airflow.api_fastapi.common.router import AirflowRouter
from airflow.providers.amazon.aws.auth_manager.constants import CONF_SAML_METADATA_URL_KEY, CONF_SECTION_NAME
from airflow.providers.amazon.aws.auth_manager.datamodels.login import LoginResponse
from airflow.providers.amazon.aws.auth_manager.user import AwsAuthManagerUser
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_1_1_PLUS
from airflow.providers.common.compat.sdk import conf

try:
    from onelogin.saml2.auth import OneLogin_Saml2_Auth
    from onelogin.saml2.errors import OneLogin_Saml2_Error
    from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
except ImportError:
    raise ImportError(
        "AWS auth manager requires the python3-saml library but it is not installed by default. "
        "Please install the python3-saml library by running: "
        "pip install apache-airflow-providers-amazon[python3-saml]"
    )

log = logging.getLogger(__name__)
login_router = AirflowRouter(tags=["AWSAuthManagerLogin"])


@login_router.get("/login")
def login(request: Request):
    """Initiate the authentication."""
    saml_auth = _init_saml_auth(request)
    callback_url = saml_auth.login("login-redirect")
    return RedirectResponse(url=callback_url)


@login_router.get("/login/token")
def login_token(request: Request) -> RedirectResponse:
    """Initiate the authentication to create a token."""
    saml_auth = _init_saml_auth(request)
    callback_url = saml_auth.login("login-token")
    return RedirectResponse(url=callback_url)


@login_router.post("/login_callback")
def login_callback(request: Request):
    """Authenticate the user."""
    saml_auth = _init_saml_auth(request)
    try:
        saml_auth.process_response()
    except OneLogin_Saml2_Error as e:
        log.exception(e)
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, "Failed to authenticate")
    errors = saml_auth.get_errors()
    is_authenticated = saml_auth.is_authenticated()
    if not is_authenticated:
        error_reason = saml_auth.get_last_error_reason()
        log.error("Failed to authenticate")
        log.error("Errors: %s", errors)
        log.error("Error reason: %s", error_reason)
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, f"Failed to authenticate: {error_reason}")

    attributes = saml_auth.get_attributes()
    user = AwsAuthManagerUser(
        user_id=attributes["id"][0],
        groups=attributes["groups"] or [],
        username=saml_auth.get_nameid(),
        email=attributes["email"][0] if "email" in attributes else None,
    )
    url = conf.get("api", "base_url", fallback="/")
    token = get_auth_manager().generate_jwt(user)

    form_data = anyio.from_thread.run(request.form)
    relay_state = form_data["RelayState"]

    if relay_state == "login-redirect":
        response = RedirectResponse(url=url, status_code=303)
        secure = bool(conf.get("api", "ssl_cert", fallback=""))
        # In Airflow 3.1.1 authentication changes, front-end no longer handle the token
        # See https://github.com/apache/airflow/pull/55506
        if AIRFLOW_V_3_1_1_PLUS:
            response.set_cookie(COOKIE_NAME_JWT_TOKEN, token, secure=secure, httponly=True)
        else:
            response.set_cookie(COOKIE_NAME_JWT_TOKEN, token, secure=secure)
        return response
    if relay_state == "login-token":
        return LoginResponse(access_token=token)
    raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, f"Invalid relay state: {relay_state}")


def _init_saml_auth(request: Request) -> OneLogin_Saml2_Auth:
    request_data = _prepare_request(request)
    base_url = conf.get(section="api", key="base_url")
    settings = {
        # We want to keep this flag on in case of errors.
        # It provides an error reasons, if turned off, it does not
        "debug": True,
        "sp": {
            "entityId": "aws-auth-manager-saml-client",
            "assertionConsumerService": {
                "url": f"{base_url.rstrip('/')}{AUTH_MANAGER_FASTAPI_APP_PREFIX}/login_callback",
                "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
            },
        },
    }
    merged_settings = OneLogin_Saml2_IdPMetadataParser.merge_settings(_get_idp_data(), settings)
    return OneLogin_Saml2_Auth(request_data, merged_settings)


def _prepare_request(request: Request) -> dict:
    host = request.headers.get("host", request.client.host if request.client else "localhost")
    data: dict[str, Any] = {
        "https": "on" if request.url.scheme == "https" else "off",
        "http_host": host,
        "server_port": request.url.port,
        "script_name": request.url.path,
        "get_data": request.query_params,
        "post_data": {},
    }
    form_data = anyio.from_thread.run(request.form)
    if "SAMLResponse" in form_data:
        data["post_data"]["SAMLResponse"] = form_data["SAMLResponse"]
    if "RelayState" in form_data:
        data["post_data"]["RelayState"] = form_data["RelayState"]
    return data


def _get_idp_data() -> dict:
    saml_metadata_url = conf.get_mandatory_value(CONF_SECTION_NAME, CONF_SAML_METADATA_URL_KEY)
    return OneLogin_Saml2_IdPMetadataParser.parse_remote(saml_metadata_url)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.notifications.sns import SnsNotifier, send_sns_notification
from airflow.providers.amazon.version_compat import NOTSET

PUBLISH_KWARGS = {
    "target_arn": "arn:aws:sns:us-west-2:123456789098:TopicName",
    "message": "foo-bar",
    "subject": "spam-egg",
    "message_attributes": {},
}


class TestSnsNotifier:
    def test_class_and_notifier_are_same(self):
        assert send_sns_notification is SnsNotifier

    @pytest.mark.parametrize(
        "aws_conn_id",
        [
            pytest.param("aws_test_conn_id", id="custom-conn"),
            pytest.param(None, id="none-conn"),
            pytest.param(NOTSET, id="default-value"),
        ],
    )
    @pytest.mark.parametrize(
        "region_name",
        [
            pytest.param("eu-west-2", id="custom-region"),
            pytest.param(None, id="no-region"),
            pytest.param(NOTSET, id="default-value"),
        ],
    )
    def test_parameters_propagate_to_hook(self, aws_conn_id, region_name):
        """Test notifier attributes propagate to SnsHook."""

        notifier_kwargs = {}
        if aws_conn_id is not NOTSET:
            notifier_kwargs["aws_conn_id"] = aws_conn_id
        if region_name is not NOTSET:
            notifier_kwargs["region_name"] = region_name

        notifier = SnsNotifier(**notifier_kwargs, **PUBLISH_KWARGS)
        with mock.patch("airflow.providers.amazon.aws.notifications.sns.SnsHook") as mock_hook:
            hook = notifier.hook
            assert hook is notifier.hook, "Hook property not cached"
            mock_hook.assert_called_once_with(
                aws_conn_id=(aws_conn_id if aws_conn_id is not NOTSET else "aws_default"),
                region_name=(region_name if region_name is not NOTSET else None),
            )

            # Basic check for notifier
            notifier.notify({})
            mock_hook.return_value.publish_to_target.assert_called_once_with(**PUBLISH_KWARGS)

    @pytest.mark.asyncio
    async def test_async_notify(self):
        notifier = SnsNotifier(**PUBLISH_KWARGS)
        with mock.patch("airflow.providers.amazon.aws.notifications.sns.SnsHook") as mock_hook:
            mock_hook.return_value.apublish_to_target = mock.AsyncMock()

            await notifier.async_notify({})

            mock_hook.return_value.apublish_to_target.assert_called_once_with(**PUBLISH_KWARGS)

    def test_sns_notifier_templated(self, create_dag_without_db):
        notifier = SnsNotifier(
            aws_conn_id="{{ dag.dag_id }}",
            target_arn="arn:aws:sns:{{ var_region }}:{{ var_account }}:{{ var_topic }}",
            message="I, {{ var_username }}",
            subject="{{ var_subject }}",
            message_attributes={"foo": "{{ dag.dag_id }}"},
            region_name="{{ var_region }}",
        )
        with mock.patch("airflow.providers.amazon.aws.notifications.sns.SnsHook") as m:
            notifier(
                {
                    "dag": create_dag_without_db("test_sns_notifier_templated"),
                    "var_username": "Robot",
                    "var_region": "us-west-1",
                    "var_account": "000000000000",
                    "var_topic": "AwesomeTopic",
                    "var_subject": "spam-egg",
                }
            )
            # Hook initialisation
            m.assert_called_once_with(aws_conn_id="test_sns_notifier_templated", region_name="us-west-1")
            # Publish message
            m.return_value.publish_to_target.assert_called_once_with(
                target_arn="arn:aws:sns:us-west-1:000000000000:AwesomeTopic",
                message="I, Robot",
                subject="spam-egg",
                message_attributes={"foo": "test_sns_notifier_templated"},
            )
